#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
from pxr import Tf
Tf.PreparePythonModule()
del Tf

from pxr.Gf import TimeCode
from pxr.Vt import TimeCodeArray

def Find(layerFileName, scenePath=None):
    '''Find(layerFileName, scenePath) -> object

layerFileName: string
scenePath: Path

If given a single string argument, returns the layer with 
the given filename.  If given two arguments (a string and a Path), finds 
the layer with the given filename and returns the scene object 
within it at the given path.'''
    layer = Layer.Find(layerFileName)
    if (scenePath is None): return layer
    return layer.GetObjectAtPath(scenePath)


# Test utilities
def _PathElemsToPrefixes(absolute, elements, numPrefixes=0):

    prefixes = []

    string = "/" if absolute else ""
        
    lastElemWasDotDot = False
    didFirst = False

    for elem in elements:
        if elem == Path.parentPathElement:
            # dotdot
            if didFirst:
                string = string + "/"
            else:
                didFirst = True
            string = string + elem
            prefixes.append(Path(string))
            lastElemWasDotDot = True
        elif elem[0] == ".":
            # property
            if lastElemWasDotDot:
                string = string + "/"
            string = string + elem
            prefixes.append(Path(string))
            lastElemWasDotDot = False
        elif elem[0] == "[":
            # rel attr or sub-attr indices, don't care which
            string = string + elem
            prefixes.append(Path(string))
            lastElemWasDotDot = False
        else:
            if didFirst:
                string = string + "/"
            else:
                didFirst = True
            string = string + elem
            prefixes.append(Path(string))
            lastElemWasDotDot = False

    if not string:
        return []
    
    return prefixes if numPrefixes == 0 else prefixes[-numPrefixes:]



#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

# pylint: disable=dict-keys-not-iterating

from __future__ import print_function

from pxr import Tf

from .common import DefaultFontFamily
from .qt import QtCore, QtGui, QtWidgets
from .usdviewApi import UsdviewApi

from code import InteractiveInterpreter
import os, sys, keyword

# just a handy debugging method
def _PrintToErr(line):
    old = sys.stdout
    sys.stdout = sys.__stderr__
    print(line)
    sys.stdout = old

def _Redirected(method):
    def new(self, *args, **kw):
        old = sys.stdin, sys.stdout, sys.stderr
        sys.stdin, sys.stdout, sys.stderr = self, self, self
        try:
            ret = method(self, *args, **kw)
        finally:
            sys.stdin, sys.stdout, sys.stderr = old
        return ret
    return new

class _Completer(object):
    """Taken from rlcompleter, with readline references stripped, and a local
    dictionary to use."""

    def __init__(self, locals):
        self.locals = locals

    def Complete(self, text, state):
        """Return the next possible completion for 'text'.
        This is called successively with state == 0, 1, 2, ... until it
        returns None.  The completion should begin with 'text'.
        """
        if state == 0:
            if "." in text:
                self.matches = self._AttrMatches(text)
            else:
                self.matches = self._GlobalMatches(text)
        try:
            return self.matches[state]
        except IndexError:
            return None

    def _GlobalMatches(self, text):
        """Compute matches when text is a simple name.

        Return a list of all keywords, built-in functions and names
        currently defines in __main__ that match.
        """

        import builtins
        import __main__

        matches = set()
        n = len(text)
        for l in [keyword.kwlist,builtins.__dict__.keys(),
                  __main__.__dict__.keys(), self.locals.keys()]:
            for word in l:
                if word[:n] == text and word != "__builtins__":
                    matches.add(word)
        return list(matches)

    def _AttrMatches(self, text):
        """Compute matches when text contains a dot.

        Assuming the text is of the form NAME.NAME....[NAME], and is
        evaluatable in the globals of __main__, it will be evaluated
        and its attributes (as revealed by dir()) are used as possible
        completions.  (For class instances, class members are are also
        considered.)

        WARNING: this can still invoke arbitrary C code, if an object
        with a __getattr__ hook is evaluated.
        """
        import re, __main__

        assert len(text)

        # This is all a bit hacky, but that's tab-completion for you.

        # Now find the last index in the text of a set of characters, and split
        # the string into a prefix and suffix token there.  The suffix token
        # will be used for completion.
        splitChars = ' )(;,+=*/-%!<>'
        index = -1
        for char in splitChars:
            index = max(text.rfind(char), index)

        if index >= len(text)-1:
            return []

        prefix = ''
        suffix = text
        if index >= 0:
            prefix = text[:index+1]
            suffix = text[index+1:]

        m = re.match(r"([^.]+(\.[^.]+)*)\.(.*)", suffix)
        if not m:
            return []
        expr, attr = m.group(1, 3)

        try:
            myobject = eval(expr, __main__.__dict__, self.locals)
        except (AttributeError, NameError, SyntaxError):
            return []

        words = set(dir(myobject))
        if hasattr(myobject,'__class__'):
            words.add('__class__')
        words = words.union(set(_GetClassMembers(myobject.__class__)))

        words = list(words)
        matches = set()
        n = len(attr)
        for word in words:
            if word[:n] == attr and word != "__builtins__":
                matches.add("%s%s.%s" % (prefix, expr, word))
        return list(matches)

def _GetClassMembers(cls):
    ret = dir(cls)
    if hasattr(cls, '__bases__'):
        for base in cls.__bases__:
            ret = ret + _GetClassMembers(base)
    return ret


class Interpreter(InteractiveInterpreter):
    def __init__(self, locals = None):
        InteractiveInterpreter.__init__(self,locals)
        self._outputBrush = None

    # overridden
    def showsyntaxerror(self, filename = None):
        self._outputBrush = QtGui.QBrush(QtGui.QColor('#ffcc63'))

        try:
            InteractiveInterpreter.showsyntaxerror(self, filename)
        finally:
            self._outputBrush = None

    # overridden
    def showtraceback(self):
        self._outputBrush = QtGui.QBrush(QtGui.QColor('#ff0000'))

        try:
            InteractiveInterpreter.showtraceback(self)
        finally:
            self._outputBrush = None

    def GetOutputBrush(self):
        return self._outputBrush

# Modified from site.py in the Python distribution.
#
# This allows each interpreter editor to have it's own Helper object.
# The built-in pydoc.help grabs sys.stdin and sys.stdout the first time it
# is run and then never lets them go.
class _Helper(object):
    """Define a replacement for the built-in 'help'.
    This is a wrapper around pydoc.Helper (with a twist).

    """
    def __init__(self, input, output):
        import pydoc
        self._helper = pydoc.Helper(input, output)

    def __repr__(self):
        return "Type help() for interactive help, " \
               "or help(object) for help about object."

    def __call__(self, *args, **kwds):
        return self._helper(*args, **kwds)


class Controller(QtCore.QObject):
    """
    Controller is a Python shell written using Qt.

    This class is a controller between Python and something which acts
    like a QTextEdit.

    """

    _isAnyReadlineEventLoopActive = False

    def __init__(self, textEdit, initialPrompt, locals = None):
        """Constructor.

        The optional 'locals' argument specifies the dictionary in
        which code will be executed; it defaults to a newly created
        dictionary with key "__name__" set to "__console__" and key
        "__doc__" set to None.

        """

        super(Controller, self).__init__()

        self.interpreter = Interpreter(locals)
        self.interpreter.locals['help'] = _Helper(self, self)

        self.completer = _Completer(self.interpreter.locals)
        # last line + last incomplete lines
        self.lines = []

        # flag: the interpreter needs more input to run the last lines.
        self.more = 0

        # history
        self.history = []
        self.historyPointer = None
        self.historyInput = ''

        # flag: readline() is being used for e.g. raw_input and input().
        # We use a nested QEventloop here because we want to emulate
        # modeless UI even though the readline protocol requires blocking calls.
        self.readlineEventLoop = QtCore.QEventLoop(textEdit)

        # interpreter prompt.
        try:
            sys.ps1
        except AttributeError:
            sys.ps1 = ">>> "
        try:
            sys.ps2
        except AttributeError:
            sys.ps2 = "... "

        self.textEdit = textEdit
        self.textEdit.destroyed.connect(self._TextEditDestroyedSlot)

        self.textEdit.returnPressed.connect(self._ReturnPressedSlot)

        self.textEdit.requestComplete.connect(self._CompleteSlot)

        self.textEdit.requestNext.connect(self._NextSlot)

        self.textEdit.requestPrev.connect(self._PrevSlot)

        appInstance = QtWidgets.QApplication.instance()
        appInstance.aboutToQuit.connect(self._QuitSlot)

        self.textEdit.setTabChangesFocus(False)

        self.textEdit.setWordWrapMode(QtGui.QTextOption.WrapMode.WrapAnywhere)
        self.textEdit.setWindowTitle('Interpreter')

        self.textEdit.promptLength = len(sys.ps1)

        # Do initial auto-import.
        self._DoAutoImports()

        # interpreter banner
        self.write('Python %s on %s.\n' % (sys.version, sys.platform))

        # Run $PYTHONSTARTUP startup script.
        startupFile = os.getenv('PYTHONSTARTUP')
        if startupFile:
            path = os.path.realpath(os.path.expanduser(startupFile))
            if os.path.isfile(path):
                self.ExecStartupFile(path)

        self.write(initialPrompt)
        self.write(sys.ps1)
        self.SetInputStart()

    def _DoAutoImports(self):
        modules = Tf.ScriptModuleLoader().GetModulesDict()
        for name, mod in modules.items():
            self.interpreter.runsource('import ' + mod.__name__ +
                                       ' as ' + name + '\n')

    @_Redirected
    def ExecStartupFile(self, path):
        # fix for bug 9104
        # this sets __file__ in the globals dict while we are execing these
        # various startup scripts, so that they can access the location from
        # which they are being run.
        # also, update the globals dict after we exec the file (bug 9529)
        self.interpreter.runsource( 
            'g = dict(globals());' 
            'g["__file__"] = "{0}";'
            'f = open("{0}", "rb");'
            'exec(compile(f.read(), "{0}", "exec"), g);'
            'f.close();'
            'del g["__file__"];'
            'globals().update(g);'.format(path))
        self.SetInputStart()
        self.lines = []

    def SetInputStart(self):
        cursor = self.textEdit.textCursor()
        cursor.movePosition(QtGui.QTextCursor.MoveOperation.End)
        self.textEdit.SetStartOfInput(cursor.position())

    def _QuitSlot(self):
        if self.readlineEventLoop:
            if self.readlineEventLoop.isRunning():
                self.readlineEventLoop.Exit()

    def _TextEditDestroyedSlot(self):
        self.readlineEventLoop = None

    def _ReturnPressedSlot(self):
        if self.readlineEventLoop.isRunning():
            self.readlineEventLoop.Exit()
        else:
            self._Run()

    def flush(self):
        """
        Simulate stdin, stdout, and stderr.
        """
        pass

    def isatty(self):
        """
        Simulate stdin, stdout, and stderr.
        """
        return 1

    def readline(self):
        """
        Simulate stdin, stdout, and stderr.
        """
        # XXX: Prevent more than one interpreter from blocking on a readline()
        #      call.  Starting more than one subevent loop does not work,
        #      because they must exit in the order that they were created.
        if Controller._isAnyReadlineEventLoopActive:
            raise RuntimeError("Simultaneous readline() calls in multiple "
                               "interpreters are not supported.")

        cursor = self.textEdit.textCursor()
        cursor.movePosition(QtGui.QTextCursor.MoveOperation.End)
        self.SetInputStart()
        self.textEdit.setTextCursor(cursor)

        try:
            Controller._isAnyReadlineEventLoopActive = True

            # XXX TODO - Make this suck less if possible.  We're invoking a
            # subeventloop here, which means until we return from this
            # readline we'll never get up to the main event loop.  To avoid
            # using a subeventloop, we would need to return control to the
            # main event loop in the main thread, suspending the execution of
            # the code that called into here, which also lives in the main
            # thread.  This essentially requires a
            # co-routine/continuation-based solution capable of dealing with
            # an arbitrary stack of interleaved Python/C calls (e.g. greenlet).
            self.readlineEventLoop.Exec()
        finally:
            Controller._isAnyReadlineEventLoopActive = False

        cursor.movePosition(QtGui.QTextCursor.MoveOperation.EndOfBlock,
                            QtGui.QTextCursor.MoveMode.MoveAnchor)

        cursor.setPosition(self.textEdit.StartOfInput(),
                           QtGui.QTextCursor.MoveMode.KeepAnchor)
        txt = str(cursor.selectedText())

        if len(txt) == 0:
            return '\n'
        else:
            self.write('\n')
            return txt

    @_Redirected
    def write(self, text):
        'Simulate stdin, stdout, and stderr.'


        # Move the cursor to the end of the document
        self.textEdit.moveCursor(QtGui.QTextCursor.MoveOperation.End)

        # Clear any existing text format.  We will explicitly set the format
        # later to something else if need be.
        self.textEdit.ResetCharFormat()

        # Copy the textEdit's current cursor.
        cursor = self.textEdit.textCursor()
        try:
            # If there's a designated output brush, merge that character format
            # into the cursor's character format.
            if self.interpreter.GetOutputBrush():
                cf = QtGui.QTextCharFormat()
                cf.setForeground(self.interpreter.GetOutputBrush())
                cursor.mergeCharFormat(cf)

            # Write the text to the textEdit.
            cursor.insertText(text)

        finally:
            # Set the textEdit's cursor to the end of input
            self.textEdit.moveCursor(QtGui.QTextCursor.MoveOperation.End)

    # get the length of a string in pixels bases on our current font
    @staticmethod
    def _GetStringLengthInPixels(cf, string):
        font = cf.font()
        fm = QtGui.QFontMetrics(font)
        strlen = fm.horizontalAdvance(string)
        return strlen

    def _CompleteSlot(self):

        cf = self.textEdit.currentCharFormat()

        line = self._GetInputLine()
        cursor = self.textEdit.textCursor()
        origPos = cursor.position()
        cursor.setPosition(self.textEdit.StartOfInput(),
                           QtGui.QTextCursor.MoveMode.KeepAnchor)
        text = str(cursor.selectedText())
        tokens = text.split()
        token = ''
        if len(tokens) != 0:
            token = tokens[-1]

        completions = []
        p = self.completer.Complete(token,len(completions))
        while p != None:
            completions.append(p)
            p = self.completer.Complete(token, len(completions))

        if len(completions) == 0:
            return

        elif len(completions) != 1:
            self.write("\n")

            contentsRect = self.textEdit.contentsRect()
            # get the width inside the margins to eventually determine the
            # number of columns in our text table based on the max string width
            # of our completed words XXX TODO - paging based on widget height
            width = contentsRect.right() - contentsRect.left()

            maxLength = 0
            for i in completions:
                maxLength = max(maxLength, self._GetStringLengthInPixels(cf, i))
            # pad it a bit
            maxLength = maxLength + self._GetStringLengthInPixels(cf, '  ')

            # how many columns can we fit on screen?
            numCols = max(1,width // maxLength)
            # how many rows do we need to fit our data
            numRows = (len(completions) // numCols) + 1

            columnWidth = QtGui.QTextLength(QtGui.QTextLength.Type.FixedLength,
                                            maxLength)

            tableFormat = QtGui.QTextTableFormat()
            tableFormat.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeft)
            tableFormat.setCellPadding(0)
            tableFormat.setCellSpacing(0)
            tableFormat.setColumnWidthConstraints([columnWidth] * numCols)
            tableFormat.setBorder(0)
            cursor = self.textEdit.textCursor()

            # Make the completion table insertion a single edit block
            cursor.beginEditBlock()
            cursor.movePosition(QtGui.QTextCursor.MoveOperation.End)
            textTable = cursor.insertTable(numRows, numCols, tableFormat)

            completions.sort()
            index = 0
            completionsLength = len(completions)

            for col in range(0,numCols):
                for row in range(0,numRows):
                    cellNum = (row * numCols) + col
                    if (cellNum >= completionsLength):
                        continue
                    tableCell = textTable.cellAt(row,col)
                    cellCursor = tableCell.firstCursorPosition()
                    cellCursor.insertText(completions[index], cf)
                    index +=1

            cursor.endEditBlock()

            self.textEdit.setTextCursor(cursor)
            self.write("\n")
            if self.more:
                self.write(sys.ps2)
            else:
                self.write(sys.ps1)

            self.SetInputStart()
            # complete up to the common prefix
            cp = os.path.commonprefix(completions)

            # make sure that we keep everything after the cursor the same as it
            # was previously
            i = line.rfind(token)
            textToRight = line[i+len(token):]
            line = line[0:i] + cp + textToRight
            self.write(line)

            # replace the line and reset the cursor
            cursor = self.textEdit.textCursor()
            cursor.setPosition(self.textEdit.StartOfInput() + len(line) -
                               len(textToRight))

            self.textEdit.setTextCursor(cursor)

        else:
            i = line.rfind(token)
            line = line[0:i] + completions[0] + line[i+len(token):]

            # replace the line and reset the cursor
            cursor = self.textEdit.textCursor()

            cursor.setPosition(self.textEdit.StartOfInput(),
                               QtGui.QTextCursor.MoveMode.MoveAnchor)

            cursor.movePosition(QtGui.QTextCursor.MoveOperation.EndOfBlock,
                                QtGui.QTextCursor.MoveMode.KeepAnchor)

            cursor.removeSelectedText()
            cursor.insertText(line)
            cursor.setPosition(origPos + len(completions[0]) - len(token))
            self.textEdit.setTextCursor(cursor)

    def _NextSlot(self):
        if len(self.history):
            # if we have no history pointer, we can't go forward..
            if (self.historyPointer == None):
                return
            # if we are at the end of our history stack, we can't go forward
            elif (self.historyPointer == len(self.history) - 1):
                self._ClearLine()
                self.write(self.historyInput)
                self.historyPointer = None
                return
            self.historyPointer += 1
            self._Recall()

    def _PrevSlot(self):
        if len(self.history):
            # if we have no history pointer, set it to the most recent
            # item in the history stack, and stash away our current input
            if (self.historyPointer == None):
                self.historyPointer = len(self.history)
                self.historyInput = self._GetInputLine()
            # if we are at the end of our history, beep
            elif (self.historyPointer <= 0):
                return
            self.historyPointer -= 1
            self._Recall()

    def _IsBlank(self, txt):
        return len(txt.strip()) == 0

    def _GetInputLine(self):
        cursor = self.textEdit.textCursor()
        cursor.setPosition(self.textEdit.StartOfInput(),
                           QtGui.QTextCursor.MoveMode.MoveAnchor)
        cursor.movePosition(QtGui.QTextCursor.MoveOperation.EndOfBlock,
                            QtGui.QTextCursor.MoveMode.KeepAnchor)
        txt = str(cursor.selectedText())
        return txt

    def _ClearLine(self):
        cursor = self.textEdit.textCursor()

        cursor.setPosition(self.textEdit.StartOfInput(),
                           QtGui.QTextCursor.MoveMode.MoveAnchor)

        cursor.movePosition(QtGui.QTextCursor.MoveOperation.EndOfBlock,
                            QtGui.QTextCursor.MoveMode.KeepAnchor)

        cursor.removeSelectedText()

    @_Redirected
    def _Run(self):
        """
        Append the last line to the history list, let the interpreter execute
        the last line(s), and clean up accounting for the interpreter results:
        (1) the interpreter succeeds
        (2) the interpreter fails, finds no errors and wants more line(s)
        (3) the interpreter fails, finds errors and writes them to sys.stderr
        """
        self.historyPointer = None
        inputLine = self._GetInputLine()
        if (inputLine != ""):
            self.history.append(inputLine)

        self.lines.append(inputLine)
        source = '\n'.join(self.lines)
        self.write('\n')
        self.more = self.interpreter.runsource(source)
        if self.more:
            self.write(sys.ps2)
            self.SetInputStart()
        else:
            self.write(sys.ps1)
            self.SetInputStart()
            self.lines = []

    def _Recall(self):
        """
        Display the current item from the command history.
        """
        self._ClearLine()
        self.write(self.history[self.historyPointer])

class View(QtWidgets.QTextEdit):
    """View is a QTextEdit which provides some extra
    facilities to help implement an interpreter console.  In particular,
    QTextEdit does not provide for complete control over the buffer being
    edited.  Some signals are emitted *after* action has already been
    taken, disallowing controller classes from really controlling the widget.
    This widget fixes that.
    """

    returnPressed = QtCore.Signal()
    requestPrev = QtCore.Signal()
    requestNext = QtCore.Signal()
    requestComplete = QtCore.Signal()

    def __init__(self, parent=None):
        super(View, self).__init__(parent)
        self.promptLength = 0
        self.__startOfInput = 0
        self.setUndoRedoEnabled(False)
        self.setAcceptRichText(False)
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.NoContextMenu)
        self.tripleClickTimer = QtCore.QBasicTimer()
        self.tripleClickPoint = QtCore.QPoint()
        self.ResetCharFormat()

    def SetStartOfInput(self, position):
        self.__startOfInput = position

    def StartOfInput(self):
        return self.__startOfInput

    def ResetCharFormat(self):
        charFormat = QtGui.QTextCharFormat()
        charFormat.setFontFamily(DefaultFontFamily.MONOSPACE_FONT_FAMILY)
        self.setCurrentCharFormat(charFormat)

    def _PositionInInputArea(self, position):
        return position - self.__startOfInput

    def _PositionIsInInputArea(self, position):
        return self._PositionInInputArea(position) >= 0

    def _CursorIsInInputArea(self):
        return self._PositionIsInInputArea(self.textCursor().position())

    def _SelectionIsInInputArea(self):
        if (not self.textCursor().hasSelection()):
            return False
        selStart = self.textCursor().selectionStart()
        selEnd = self.textCursor().selectionEnd()
        return self._PositionIsInInputArea(selStart) and \
               self._PositionIsInInputArea(selEnd)

    def _MoveCursorToStartOfInput(self, select=False):
        cursor = self.textCursor()
        anchor = QtGui.QTextCursor.MoveMode.MoveAnchor

        if (select):
            anchor = QtGui.QTextCursor.MoveMode.KeepAnchor

        cursor.movePosition(QtGui.QTextCursor.MoveOperation.End, anchor)

        cursor.setPosition(self.__startOfInput, anchor)

        self.setTextCursor(cursor)

    def _MoveCursorToEndOfInput(self, select=False):
        c = self.textCursor()
        anchor = QtGui.QTextCursor.MoveMode.MoveAnchor
        if (select):
            anchor = QtGui.QTextCursor.MoveMode.KeepAnchor

        c.movePosition(QtGui.QTextCursor.MoveOperation.End, anchor)
        self.setTextCursor(c)

    def _WritableCharsToLeftOfCursor(self):
        return (self._PositionInInputArea(self.textCursor().position()) > 0)

    def mousePressEvent(self, e):
        app = QtWidgets.QApplication.instance()

        # is this a triple click?
        if ((e.button() & QtCore.Qt.MouseButton.LeftButton) and
             self.tripleClickTimer.isActive() and
             (e.globalPos() - self.tripleClickPoint).manhattanLength() <
              app.startDragDistance() ):

            # instead of duplicating the triple click code completely, we just
            # pass it along. but we modify the selection that comes out of it
            # to exclude the prompt, if appropriate
            super(View, self).mousePressEvent(e)

            if (self._CursorIsInInputArea()):
                selStart = self.textCursor().selectionStart()
                selEnd = self.textCursor().selectionEnd()

                if (self._PositionInInputArea(selStart) < 0):
                    # remove selection up until start of input
                     self._MoveCursorToStartOfInput(False)
                     cursor = self.textCursor()
                     cursor.setPosition(selEnd, QtGui.QTextCursor.MoveMode.KeepAnchor)
                     self.setTextCursor(cursor)
        else:
            super(View, self).mousePressEvent(e)

    def mouseDoubleClickEvent(self, e):
        super(View, self).mouseDoubleClickEvent(e)
        app = QtWidgets.QApplication.instance()
        self.tripleClickTimer.start(app.doubleClickInterval(), self)
        # make a copy here, otherwise tripleClickPoint will always = globalPos
        self.tripleClickPoint = QtCore.QPoint(e.globalPos())

    def timerEvent(self, e):
        if (e.timerId() == self.tripleClickTimer.timerId()):
            self.tripleClickTimer.stop()
        else:
            super(View, self).timerEvent(e)

    def insertFromMimeData(self, source):
        if not self._CursorIsInInputArea():
            self._MoveCursorToEndOfInput()

        if source.hasText():
            text = source.text().replace('\r', '')
            textLines = text.split('\n')
            if (textLines[-1] == ''):
                textLines = textLines[:-1]
            for i in range(len(textLines)):
                line = textLines[i]
                cursor = self.textCursor()
                cursor.movePosition(QtGui.QTextCursor.MoveOperation.End)        
                cursor.insertText(line)
                cursor.movePosition(QtGui.QTextCursor.MoveOperation.End)        
                self.setTextCursor(cursor)
                if i < len(textLines) - 1:
                    self.returnPressed.emit()

    def keyPressEvent(self, e):
        """
        Handle user input a key at a time.
        """

        key = e.key()

        ctrl = e.modifiers() & QtCore.Qt.KeyboardModifier.ControlModifier
        alt = e.modifiers() & QtCore.Qt.KeyboardModifier.AltModifier
        shift = e.modifiers() & QtCore.Qt.KeyboardModifier.ShiftModifier

        cursorInInput = self._CursorIsInInputArea()
        selectionInInput = self._SelectionIsInInputArea()
        hasSelection = self.textCursor().hasSelection()
        canBackspace = self._WritableCharsToLeftOfCursor()
        canEraseSelection = selectionInInput and cursorInInput
        if key == QtCore.Qt.Key.Key_Backspace:
            if (canBackspace and not hasSelection) or canEraseSelection:
                super(View, self).keyPressEvent(e)
        elif key == QtCore.Qt.Key.Key_Delete:
            if (cursorInInput and not hasSelection) or canEraseSelection:
                super(View, self).keyPressEvent(e)
        elif key == QtCore.Qt.Key.Key_Left:
            pos = self._PositionInInputArea(self.textCursor().position())
            if pos == 0:
                e.ignore()
            else:
                super(View, self).keyPressEvent(e)
        elif key == QtCore.Qt.Key.Key_Right:
            super(View, self).keyPressEvent(e)
        elif key == QtCore.Qt.Key.Key_Return or key == QtCore.Qt.Key.Key_Enter:
            # move cursor to end of line.
            # emit signal to tell controller enter was pressed.
            if not cursorInInput:
                self._MoveCursorToStartOfInput(False)
            cursor = self.textCursor()
            cursor.movePosition(QtGui.QTextCursor.MoveOperation.EndOfBlock)
            self.setTextCursor(cursor)
            # emit returnPressed
            self.returnPressed.emit()

        elif (key == QtCore.Qt.Key.Key_Up
                or key == QtCore.Qt.Key.Key_Down
                # support Ctrl+P and Ctrl+N for history
                # navigation along with arrows
                or (ctrl and (key == QtCore.Qt.Key.Key_P
                              or key == QtCore.Qt.Key.Key_N))
                # support Ctrl+E/End and Ctrl+A/Home for terminal
                # style nav. to the ends of the line
                or (ctrl and (key == QtCore.Qt.Key.Key_A
                              or key == QtCore.Qt.Key.Key_E))
                or (key == QtCore.Qt.Key.Key_Home
                    or key == QtCore.Qt.Key.Key_End)):
            if cursorInInput:
                if (key == QtCore.Qt.Key.Key_Up or key == QtCore.Qt.Key.Key_P):
                    self.requestPrev.emit()
                if (key == QtCore.Qt.Key.Key_Down or key == QtCore.Qt.Key.Key_N):
                    self.requestNext.emit()
                if (key == QtCore.Qt.Key.Key_A or key == QtCore.Qt.Key.Key_Home):
                    self._MoveCursorToStartOfInput(select=shift)
                if (key == QtCore.Qt.Key.Key_E or key == QtCore.Qt.Key.Key_End):
                    self._MoveCursorToEndOfInput(select=shift)
                e.ignore()
            else:
                super(View, self).keyPressEvent(e)
        elif key == QtCore.Qt.Key.Key_Tab:
            self.AutoComplete()
            e.accept()
        elif ((ctrl and key == QtCore.Qt.Key.Key_C) or
              (shift and key == QtCore.Qt.Key.Key_Insert)):
            # Copy should never move cursor.
            super(View, self).keyPressEvent(e)
        elif ((ctrl and key == QtCore.Qt.Key.Key_X) or
              (shift and key == QtCore.Qt.Key.Key_Delete)):
            # Disallow cut from outside the input area so users don't
            # affect the scrollback buffer.
            if not selectionInInput:
                e.ignore()
            else:
                super(View, self).keyPressEvent(e)
        elif (key == QtCore.Qt.Key.Key_Control or
              key == QtCore.Qt.Key.Key_Alt or
              key == QtCore.Qt.Key.Key_Shift):
            # Ignore modifier keypresses by themselves so the cursor
            # doesn't jump to the end of input when users begin a
            # key combination.
            e.ignore()
        else:
            # All other keypresses should append to the end of input.
            if not cursorInInput:
                self._MoveCursorToEndOfInput()
            super(View, self).keyPressEvent(e)

    def AutoComplete(self):
        if self._CursorIsInInputArea():
            self.requestComplete.emit()

    def _MoveCursorToBeginning(self, select=False):
        if self._CursorIsInInputArea():
            self._MoveCursorToStartOfInput(select)
        else:
            cursor = self.textCursor()
            anchor = QtGui.QTextCursor.MoveMode.MoveAnchor
            if (select):
                anchor = QtGui.QTextCursor.MoveMode.KeepAnchor
            cursor.setPosition(0, anchor)
            self.setTextCursor(cursor)

    def _MoveCursorToEnd(self, select=False):
        if self._CursorIsInInputArea():
            self._MoveCursorToEndOfInput(select)
        else:
            cursor = self.textCursor()
            anchor = QtGui.QTextCursor.MoveMode.MoveAnchor
            if (select):
                anchor = QtGui.QTextCursor.MoveMode.KeepAnchor

            cursor.setPosition(self.__startOfInput, anchor)
            cursor.movePosition(QtGui.QTextCursor.MoveOperation.Up, anchor)
            cursor.movePosition(QtGui.QTextCursor.MoveOperation.EndOfLine, anchor)
            self.setTextCursor(cursor)

    def MoveCursorToBeginning(self):
        self._MoveCursorToBeginning(False)

    def MoveCursorToEnd(self):
        self._MoveCursorToEnd(False)

    def SelectToTop(self):
        self._MoveCursorToBeginning(True)

    def SelectToBottom(self):
        self._MoveCursorToEnd(True)


FREQUENTLY_USED = [
    "dataModel", "stage", "frame", "prim", "property", "spec", "layer"]


INITIAL_PROMPT = """
Use the `usdviewApi` variable to interact with UsdView.
Type `help(usdviewApi)` to view available API methods and properties.

Frequently used properties:
{}\n""".format(
    "".join("    usdviewApi.{} - {}\n".format(
        name, getattr(UsdviewApi, name).__doc__)
    for name in FREQUENTLY_USED))


class Myconsole(View):

    def __init__(self, parent, usdviewApi):
        super(Myconsole, self).__init__(parent)
        self.setObjectName("Myconsole")

        # Inject the UsdviewApi into the interpreter variables.
        interpreterLocals = vars()
        interpreterLocals["usdviewApi"] = usdviewApi

        # Make a Controller.
        self._controller = Controller(self, INITIAL_PROMPT, interpreterLocals)

    def locals(self):
        return self._controller.interpreter.locals



#
# Copyright 2017 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

from collections import OrderedDict

from pxr import Sdf, Gf
from .qt import QtCore

from .customAttributes import (ComputedPropertyNames, BoundingBoxAttribute,
                               LocalToWorldXformAttribute, ComputedPropertyFactory)


# Indicates that all instances of a prim are selected.
ALL_INSTANCES = -1


class Blocker:
    """Object which can be used to temporarily block the execution of a body of
    code. This object is a context manager, and enters a 'blocked' state when
    used in a 'with' statement. The 'blocked()' method can be used to find if
    the Blocker is in this 'blocked' state.
    """

    def __init__(self, exitCallback=lambda: None):

        # A count is used rather than a 'blocked' flag to allow for nested
        # blocking.
        self._count = 0

        # Fired when the Blocker's context is exited.
        self._exitCallback = exitCallback

    def __enter__(self):
        """Enter the 'blocked' state until the context is exited."""

        self._count += 1

    def __exit__(self, *args):
        """Exit the 'blocked' state."""

        self._count -= 1

        if not self.blocked():
            self._exitCallback()

    def blocked(self):
        """Returns True if in the 'blocked' state, and False otherwise."""

        return self._count > 0


class _PrimSelection(object):
    """This class keeps track of the core data for prim selection: paths and
    instances. The methods here can be called in any order required without
    corrupting the path selection state.
    """

    def __init__(self):

        # The order of paths selected needs to be maintained so we can track the
        # focus path. An OrderedDict is more efficient than a list here since it
        # supports efficient removal of arbitrary paths but still maintains the
        # path order. Sdf.Path objects are the keys in the OrderedDict, and a
        # set of selected instances are the values. If all instances are
        # selected, None is the value rather than a set.
        self._selection = OrderedDict()

        self._added = set()
        self._removed = set()

    def _clearPrimPath(self, path):
        """Clears a path from the selection and updates the diff."""

        if path in self._added:
            # Path was added in this diff, but we are removing it again. Since
            # there is no net change, it shouldn't be in _added or _removed.
            self._added.discard(path)
        else:
            self._removed.add(path)

        del self._selection[path]

    def _discardInstance(self, path, instance):
        """Discards an instance from the selection, then deletes the path from
        the selection if it has no more instances.
        """

        instances = self._selection[path]
        instances.discard(instance)
        if len(instances) == 0:
            # Last instance deselected, so path should be deselected.
            self._clearPrimPath(path)

    def _allInstancesSelected(self, path):
        """Returns True if all instances of a specified path are selected and
        False otherwise.
        """

        if path in self._selection:
            return self._selection[path] == ALL_INSTANCES
        else:
            return False

    def _noInstancesSelected(self, path):
        """Returns True if all instances of a specified path are selected and
        False otherwise.
        """

        return path not in self._selection

    def clear(self):
        """Clear the path selection."""

        # _clearPrimPath modifies self._selection, so make a copy of keys
        # before mutating it.
        for path in list(self._selection.keys()):
            self._clearPrimPath(path)
    
    def removeMatchingPaths(self, matches):
        """Remove any paths that pass the given predicate"""

        # _clearPrimPath modifies self._selection, so make a copy of keys
        # before mutating it.
        for path in list(self._selection.keys()):
            if matches(path):
                self._clearPrimPath(path)

    def addPrimPath(self, path, instance=ALL_INSTANCES):
        """Add a path to the selection. If an instance is given, then only add
        that instance. If all instances are selected when this happens then the
        single instance will become the only selected one.
        """

        # If the path is not already in the selection, update the diff.
        if path not in self._selection:
            if path in self._removed:
                # Path was removed in this diff, but is back now. Since there is
                # no net change, it shouldn't be in _added or _removed.
                self._removed.discard(path)
            else:
                self._added.add(path)

        if instance == ALL_INSTANCES:
            # Trying to add all instances, make sure all instances are selected.
            self._selection[path] = ALL_INSTANCES
        else:
            # Trying to add a single instance.
            if self._allInstancesSelected(path) or self._noInstancesSelected(path):
                # Either all instances selected or none selected. Create an
                # empty set of instances then add the target instance.
                self._selection[path] = set()
            self._selection[path].add(instance)

    def removePrimPath(self, path, instance=ALL_INSTANCES):
        """Remove a path from the selection. If an instance is given, then only
        remove that instance. If all instances are selected when this happens,
        deselect all instances. If the target does not exist in the selection,
        do nothing.
        """

        if path in self._selection:
            if instance == ALL_INSTANCES or self._allInstancesSelected(path):
                # Want to deselect all instances or all instances are selected.
                # Either way, deselect all.
                self._clearPrimPath(path)
            else:
                # Some instances selected and want to deselect one of them.
                self._discardInstance(path, instance)

    def togglePrimPath(self, path, instance=ALL_INSTANCES):
        """Toggle the selection of a path. If an instance is given, only toggle
        that instance's selection.
        """

        if path not in self._selection:
            self.addPrimPath(path, instance)
            return

        # Path is at least partially selected.
        if instance == ALL_INSTANCES:
            # Trying to toggle all instances.
            if self._allInstancesSelected(path):
                # All instances are already selected, so deselect the path.
                self._clearPrimPath(path)
            else:
                # Only some instances are selected, select all all of them.
                self._selection[path] = ALL_INSTANCES
        else:
            # Trying to toggle a single instance.
            if self._allInstancesSelected(path):
                # Currently all instances are selected. Switch selection to
                # only the new instance.
                self._selection[path] = set([instance])
            else:
                # Some instances already selected. Toggle the new instance
                # in the selection.
                instances = self._selection[path]
                if instance in instances:
                    self._discardInstance(path, instance)
                else:
                    instances.add(instance)

    def getPrimPaths(self):
        """Get a list of paths that are at least partially selected."""

        return list(self._selection.keys())

    def getPrimPathInstances(self):
        """Get the full selection of paths and their corresponding selected
        instances.
        """

        return OrderedDict(
            (path, set(instances)) if isinstance(instances, set) else (path, instances)
            for path, instances in self._selection.items())

    def getDiff(self):
        """Get the prims added to or removed from the selection since the last
        time getDiff() was called.
        """

        diff = (self._added, self._removed)

        self._added = set()
        self._removed = set()

        return diff


class _PropSelection(object):
    """This class keeps track of the state of property selection."""

    def __init__(self):

        self._selection = OrderedDict()

    def clear(self):
        """Clears the property selection."""

        self._selection = OrderedDict()

    def addPropPath(self, primPath, propName):
        """Add a property to the selection."""

        propTuple = (primPath, propName)

        # If this property is already selected, remove it from the selection and
        # re-add it so it becomes the focus property.
        if propTuple in self._selection:
            targets = self._selection[propTuple]
            del self._selection[propTuple]
            self._selection[propTuple] = targets
        else:
            self._selection[propTuple] = set()

    def removePropPath(self, primPath, propName):
        """Remove a property from the selection."""

        propTuple = (primPath, propName)
        if propTuple in self._selection:
            del self._selection[propTuple]

    def addTarget(self, primPath, propName, target):
        """Add a target to the selection. Also add the target's property if it
        is not already in the selection.
        """

        propTuple = (primPath, propName)

        # If this property is already selected, remove it from the selection and
        # re-add it so it becomes the focus property.
        if propTuple in self._selection:
            targets = self._selection[propTuple]
            del self._selection[propTuple]
            self._selection[propTuple] = targets
        else:
            targets = self._selection.setdefault(propTuple, set())

        targets.add(target)

    def removeTarget(self, primPath, propName, target):
        """Remove a target from the selection. If the target or its property are
        not already in the selection, nothing is changed.
        """

        propTuple = (primPath, propName)
        if propTuple in self._selection:
            self._selection[propTuple].discard(target)

    def getPropPaths(self):
        """Get the list of properties."""

        return list(self._selection.keys())

    def getTargets(self):
        """Get a dictionary which maps selected properties to a set of their
        selected targets or connections.
        """

        propTargets = OrderedDict()
        for propTuple, targets in self._selection.items():
            propTargets[propTuple] = set(targets)

        return propTargets


class SelectionDataModel(QtCore.QObject):
    """Data model managing the current selection of prims and properties.
    Please note that the owner of an instance of this class is
    responsible for calling SelectionDataModel.removeUnpopulatedPrims() when
    appropriate, lest methods like getPrims() return invalid prims."""

    # Signals must be declared in the class, but each instance of the selection
    # data model has its own unique signal instances.

    # When emitted, includes two sets: one of newly selected prims, and one of
    # newly deselected prims.
    signalPrimSelectionChanged = QtCore.Signal(set, set)

    signalPropSelectionChanged = QtCore.Signal()
    signalComputedPropSelectionChanged = QtCore.Signal()

    def __init__(self, rootDataModel, _computedPropFactory=None):
        QtCore.QObject.__init__(self)

        self._rootDataModel = rootDataModel

        # _computedPropFactory may be passed explicitly for unit testing.
        if _computedPropFactory is None:
            self._computedPropFactory = ComputedPropertyFactory(
                self._rootDataModel)
        else:
            self._computedPropFactory = _computedPropFactory

        self.batchPrimChanges = Blocker(
            exitCallback=self._primSelectionChanged)
        self.batchPropChanges = Blocker(
            exitCallback=self._propSelectionChanged)
        self.batchComputedPropChanges = Blocker(
            exitCallback=self._computedPropSelectionChanged)

        self._pointSelection = Gf.Vec3f(0.0, 0.0, 0.0)

        self._primSelection = _PrimSelection()
        # The path selection should never be empty. If it ever is, we
        # immediately add the root path before returning to user control (see
        # _primSelectionChanged).
        self._primSelection.addPrimPath(Sdf.Path.absoluteRootPath)
        # Clear the prim selection diff so we don't get the absolute root in the
        # first signal from signalPrimSelectionChanged.
        self._primSelection.getDiff()

        self._lcdPathSelection = [Sdf.Path.absoluteRootPath]

        self._propSelection = _PropSelection()
        self._computedPropSelection = _PropSelection()

    ### Internal Operations ###
    def _primSelectionChanged(self, emitSelChangedSignal=True):
        """Should be called whenever a change is made to _primSelection. Some
        final work is done then the prim selection changed signal is emitted.
        """

        # If updates are suppressed, do not emit a signal or do any
        # pre-processing.
        if self.batchPrimChanges.blocked():
            return

        # Make sure there is always at least one path selected.
        if len(self.getPrimPaths()) == 0:
            self._primSelection.addPrimPath(Sdf.Path.absoluteRootPath)

        # Recalculate the LCD prims whenever the path selection changes.
        paths = self._primSelection.getPrimPaths()
        if len(paths) > 1:
            paths = [path for path in paths
                        if path != Sdf.Path.absoluteRootPath]
        self._lcdPathSelection = Sdf.Path.RemoveDescendentPaths(paths)

        # Finally, emit the changed signal.
        added, removed = self._primSelection.getDiff()
        if emitSelChangedSignal:
            self.signalPrimSelectionChanged.emit(added, removed)

    def _propSelectionChanged(self):
        """Should be called whenever a change is made to _propSelection."""

        # If updates are suppressed, do not emit a signal or do any
        # pre-processing.
        if self.batchPropChanges.blocked():
            return

        self.signalPropSelectionChanged.emit()

    def _computedPropSelectionChanged(self):
        """Should be called whenever a change is made to _computedPropSelection.
        """

        # If updates are suppressed, do not emit a signal or do any
        # pre-processing.
        if self.batchComputedPropChanges.blocked():
            return

        self.signalComputedPropSelectionChanged.emit()

    def _ensureValidPrimPath(self, path):
        """Validate an input path. If it is a string path, convert it to an
        Sdf.Path object.
        """

        sdfPath = Sdf.Path(str(path))
        if not sdfPath.IsAbsoluteRootOrPrimPath():
            raise ValueError("Path must be a prim path, got: {}".format(
                repr(sdfPath)))
        return sdfPath

    def _validateInstanceIndexParameter(self, instance):
        """Validate an instance used as a parameter. This can be any positive
        int or ALL_INSTANCES."""

        validIndex = False
        if isinstance(instance, int):
            if instance >= 0 or instance == ALL_INSTANCES:
                validIndex = True

        if not validIndex:
            raise ValueError(
                "Instance must be a positive int or ALL_INSTANCES"
                ", got: {}".format(repr(instance)))

    def _ensureValidPropPath(self, prop):
        """Validate a property."""

        sdfPath = Sdf.Path(str(prop))
        if not sdfPath.IsPropertyPath():
            raise ValueError("Path must be a property path, got: {}".format(
                repr(sdfPath)))
        return sdfPath

    def _ensureValidTargetPath(self, targetPath):
        """Validate a property target or connection."""
        sdfPath = Sdf.Path(str(targetPath))
        if not targetPath.IsPrimPath() and not targetPath.IsPropertyPath():
            raise ValueError("Path must be a prim or property path, got: {}".format(
                repr(sdfPath)))

        return sdfPath

    def _getPropFromPath(self, path):
        """Get a Usd property object from a property path."""

        prim = self._rootDataModel.stage.GetPrimAtPath(path.GetPrimPath())
        return prim.GetProperty(path.name)

    def _getTargetFromPath(self, path):
        """Get the Usd object from a target path. It can be either a Usd prim or
        Usd property.
        """

        if path.IsPropertyPath():
            return self._getPropFromPath(path)
        else:
            return self._rootDataModel.stage.GetPrimAtPath(path)

    def _requireNotBatchingPrims(self):
        """Raise an error if we are currently batching prim selection changes.
        We don't want to allow reading prim selection state in the middle of a
        batch.
        """

        if self.batchPrimChanges.blocked():
            raise RuntimeError(
                "Cannot get prim selection state while batching changes.")

    def _requireNotBatchingProps(self):
        """Raise an error if we are currently batching prop selection changes.
        We don't want to allow reading prop selection state in the middle of a
        batch.
        """

        if self.batchPropChanges.blocked():
            raise RuntimeError(
                "Cannot get property selection state while batching changes.")

    def _getComputedPropFromPath(self, primPath, propName):
        """Get a CustomAttribute object from a prim path and property name.
        Raise an error if the property name does not match any known
        CustomAttribute.
        """

        prim = self._rootDataModel.stage.GetPrimAtPath(primPath)
        return self._computedPropFactory.getComputedProperty(prim, propName)

    def _requireNotBatchingComputedProps(self):
        """Raise an error if we are currently batching prop selection changes.
        We don't want to allow reading prop selection state in the middle of a
        batch.
        """

        if self.batchComputedPropChanges.blocked():
            raise RuntimeError("Cannot get computed property selection state "
                "while batching changes.")

    def _buildPropPath(self, primPath, propName):
        """Build a new property path from a prim path and a property name."""

        return Sdf.Path(str(primPath) + "." + propName)

    def _validateComputedPropName(self, propName):
        """Validate a computed property name."""

        if propName not in ComputedPropertyNames:
            raise ValueError("Invalid computed property name: {}".format(
                repr(propName)))

    def _switchProps(self, fromPrimPath, toPrimPath):
        """Switch all selected properties from one prim to another. Only do this
        if all properties currently belong to the "from" prim.
        """

        propTargets = self.getPropTargetPaths()
        computedProps = self.getComputedPropPaths()

        # Check that all properties belong to the "from" prim.
        for propPath in propTargets:
            if propPath.GetPrimPath() != fromPrimPath:
                return
        for propPrimPath, propName in computedProps:
            if propPrimPath != fromPrimPath:
                return

        # Switch all properties to the "to" prim.
        with self.batchPropChanges:

            self.clearProps()

            # Root prim cannot have non-computed properties. The property paths
            # in this case are invalid.
            if str(toPrimPath) != "/":
                for propPath, targets in propTargets.items():
                    newPropPath = self._buildPropPath(toPrimPath, propPath.name)
                    self.addPropPath(newPropPath)
                    for target in targets:
                        self.addPropTargetPath(newPropPath, target)

        with self.batchComputedPropChanges:

            self.clearComputedProps()

            for primPath, propName in computedProps:
                self.addComputedPropPath(toPrimPath, propName)

    ### General Operations ###

    def clear(self):
        """Clear all selections."""
        self.clearPoint()
        self.clearPrims()
        self.clearProps()

    def clearPoint(self):
        self.setPoint(Gf.Vec3f(0.0, 0.0, 0.0))

    def setPoint(self, point):
        self._pointSelection = point

    def getPoint(self):
        return self._pointSelection

    ### Prim Path Operations ###

    def clearPrims(self):
        """Clear the prim selection (same as path selection)."""

        self._primSelection.clear()
        self._primSelectionChanged()

    def addPrimPath(self, path, instance=ALL_INSTANCES):
        """Add a path to the path selection. If an instance is given, only add
        that instance.
        """
        path = self._ensureValidPrimPath(path)
        self._validateInstanceIndexParameter(instance)

        self._primSelection.addPrimPath(path, instance)
        self._primSelectionChanged()

    def removePrimPath(self, path, instance=ALL_INSTANCES):
        """Remove a path from the path selection. If an instance is given, only
        remove that instance. If the target does not exist in the selection, do
        nothing.
        """

        path = self._ensureValidPrimPath(path)
        self._validateInstanceIndexParameter(instance)

        self._primSelection.removePrimPath(path, instance)
        self._primSelectionChanged()

    def togglePrimPath(self, path, instance=ALL_INSTANCES):
        """Toggle a path in the path selection. If an instance is given, only
        that instance is toggled.
        """

        path = self._ensureValidPrimPath(path)
        self._validateInstanceIndexParameter(instance)

        self._primSelection.togglePrimPath(path, instance)
        self._primSelectionChanged()

    def setPrimPath(self, path, instance=ALL_INSTANCES):
        """Clear the prim selection then add a single prim path back to the
        selection. If an instance is given, only add that instance.
        """
        with self.batchPrimChanges:
            self.clearPrims()
            self.addPrimPath(path, instance)

    def getFocusPrimPath(self):
        """Get the path currently in focus."""

        self._requireNotBatchingPrims()
        return self._primSelection.getPrimPaths()[0]

    def getPrimPaths(self):
        """Get a list of all selected paths."""

        self._requireNotBatchingPrims()
        return self._primSelection.getPrimPaths()

    def getLCDPaths(self):
        """Get a list of paths from the selection who do not have an ancestor
        that is also in the selection. The "Least Common Denominator" paths.
        """

        self._requireNotBatchingPrims()
        return list(self._lcdPathSelection)

    def getPrimPathInstances(self):
        """Get a dictionary which maps each selected prim to a set of its
        selected instances. If all of a path's instances are selected, the value
        is ALL_INSTANCES rather than a set.
        """

        self._requireNotBatchingPrims()
        return self._primSelection.getPrimPathInstances()

    def switchToPrimPath(self, path, instance=ALL_INSTANCES):
        """Select only the given prim path. If only a single prim was selected
        before and all selected properties belong to this prim, select the
        corresponding properties on the new prim instead. If an instance is
        given, only select that instance.
        """

        path = self._ensureValidPrimPath(path)

        oldPrimPaths = self.getPrimPaths()

        with self.batchPrimChanges:
            self.clearPrims()
            self.addPrimPath(path, instance)

        if len(oldPrimPaths) == 1:
            self._switchProps(oldPrimPaths[0], path)

    ### Prim Operations ###
    # These are all convenience methods which just call their respective path
    # operations.

    def addPrim(self, prim, instance=ALL_INSTANCES):
        """Add a prim's path to the path selection. If an instance is given,
        only add that instance.
        """

        self.addPrimPath(prim.GetPath(), instance)

    def removePrim(self, prim, instance=ALL_INSTANCES):
        """Remove a prim from the prim selection. If an instance is given, only
        remove that instance. If the target does not exist in the selection, do
        nothing.
        """

        self.removePrimPath(prim.GetPath(), instance)

    def togglePrim(self, prim, instance=ALL_INSTANCES):
        """Toggle a prim's path in the path selection. If an instance is given,
        only that instance is toggled.
        """

        self.togglePrimPath(prim.GetPath(), instance)

    def setPrim(self, prim, instance=ALL_INSTANCES):
        """Clear the prim selection then add a single prim back to the
        selection. If an instance is given, only add that instance.
        """

        self.setPrimPath(prim.GetPath(), instance)

    def getFocusPrim(self):
        """Get the prim whose path is currently in focus."""

        return self._rootDataModel.stage.GetPrimAtPath(self.getFocusPrimPath())

    def getPrims(self):
        """Get a list of all prims whose paths are selected."""

        return [self._rootDataModel.stage.GetPrimAtPath(path)
            for path in self.getPrimPaths()]

    def getLCDPrims(self):
        """Get a list of prims whose paths are both selected and do not have an
        ancestor that is also in the selection. The "Least Common Denominator"
        prims.
        """

        return [self._rootDataModel.stage.GetPrimAtPath(path)
            for path in self.getLCDPaths()]

    def getPrimInstances(self):
        """Get a dictionary which maps each prim whose path is selected to a set
        of its selected instances. If all of a path's instances are selected,
        the value is ALL_INSTANCES rather than a set.
        """

        return OrderedDict(
            (self._rootDataModel.stage.GetPrimAtPath(path), instance)
            for path, instance in self.getPrimPathInstances().items())

    def switchToPrim(self, prim, instance=ALL_INSTANCES):
        """Select only the given prim. If only a single prim was selected before
        and all selected properties belong to this prim, select the
        corresponding properties on the new prim instead.
        """

        self.switchToPrimPath(prim.GetPath(), instance)

    ### Prim Group Removal Operations ###
    # Convenience methods for removing groups of prims 
    
    def removeInactivePrims(self):
        """Remove all inactive prims"""
        for prim in self.getPrims():
            if not prim.IsActive():
                self.removePrim(prim)
    
    def removePrototypePrims(self):
        """Remove all prototype prims"""
        for prim in self.getPrims():
            if prim.IsPrototype() or prim.IsInPrototype():
                self.removePrim(prim)

    def removeAbstractPrims(self):
        """Remove all abstract prims"""
        for prim in self.getPrims():
            if prim.IsAbstract():
                self.removePrim(prim)
    
    def removeUndefinedPrims(self):
        """Remove all undefined prims"""
        for prim in self.getPrims():
            if not prim.IsDefined():
                self.removePrim(prim)

    def removeUnpopulatedPrims(self):
        """Remove all prim paths whose corresponding prims do not currently
        exist on the stage.  It is the application's responsibility to
        call this method while it is processing changes to the stage,
        *before* querying this object for selections.  Because this is a
        synchronization operation rather than an expression of GUI state
        change, it does *not* perform any notifications/signals, which could
        cause reentrant application change processing."""
        stage = self._rootDataModel.stage
        self._primSelection.removeMatchingPaths(lambda path: not stage.GetPrimAtPath(path))
        self._primSelectionChanged(emitSelChangedSignal=False)
     
    ### Property Path Operations ###

    def clearProps(self):
        """Clear the property selection."""

        self._propSelection.clear()
        self._propSelectionChanged()

    def addPropPath(self, path):
        """Add a property to the selection."""

        path = self._ensureValidPropPath(path)

        primPath = path.GetPrimPath()
        propName = path.name

        self._propSelection.addPropPath(primPath, propName)
        self._propSelectionChanged()

    def removePropPath(self, path):
        """Remove a property from the selection."""

        path = self._ensureValidPropPath(path)

        primPath = path.GetPrimPath()
        propName = path.name

        self._propSelection.removePropPath(primPath, propName)
        self._propSelectionChanged()

    def setPropPath(self, path):
        """Clear the property selection, then add a single property path back to
        the selection.
        """

        path = self._ensureValidPropPath(path)

        with self.batchPropChanges:
            self.clearProps()
            self.addPropPath(path)

    def addPropTargetPath(self, path, targetPath):
        """Select a property's target or connection."""

        path = self._ensureValidPropPath(path)
        targetPath = self._ensureValidTargetPath(targetPath)

        primPath = path.GetPrimPath()
        propName = path.name

        self._propSelection.addTarget(primPath, propName, targetPath)
        self._propSelectionChanged()

    def removePropTargetPath(self, path, targetPath):
        """Deselect a property's target or connection."""

        path = self._ensureValidPropPath(path)
        targetPath = self._ensureValidTargetPath(targetPath)

        primPath = path.GetPrimPath()
        propName = path.name

        self._propSelection.removeTarget(primPath, propName, targetPath)
        self._propSelectionChanged()

    def setPropTargetPath(self, path, targetPath):
        """Clear the property selection, then add a single property path back to
        the selection with a target.
        """

        with self.batchPropChanges:
            self.clearProps()
            self.addPropTargetPath(path, targetPath)

    def getFocusPropPath(self):
        """Get the focus property from the property selection."""

        self._requireNotBatchingProps()

        propPaths = [self._buildPropPath(*propTuple)
            for propTuple in self._propSelection.getPropPaths()]
        if len(propPaths) > 0:
            return propPaths[-1]
        else:
            return None

    def getPropPaths(self):
        """Get a list of all selected properties."""

        self._requireNotBatchingProps()

        propPaths = [self._buildPropPath(*propTuple)
            for propTuple in self._propSelection.getPropPaths()]
        return propPaths

    def getPropTargetPaths(self):
        """Get a dictionary which maps selected properties to a set of their
        selected targets or connections.
        """

        self._requireNotBatchingProps()

        return OrderedDict((self._buildPropPath(*propTuple), set(targets))
            for propTuple, targets in self._propSelection.getTargets().items())

    ### Property Operations ###

    def addProp(self, prop):
        """Add a property to the selection."""

        self.addPropPath(prop.GetPath())

    def removeProp(self, prop):
        """Remove a property from the selection."""

        self.removePropPath(prop.GetPath())

    def setProp(self, prop):
        """Clear the property selection, then add a single property back to the
        selection.
        """

        self.setPropPath(prop.GetPath())

    def addPropTarget(self, prop, target):
        """Select a property's target or connection."""

        self.addPropTargetPath(prop.GetPath(), target.GetPath())

    def removePropTarget(self, prop, target):
        """Deselect a property's target or connection."""

        self.removePropTargetPath(prop.GetPath(), target.GetPath())

    def setPropTarget(self, prop, target):
        """Clear the property selection, then add a single property back to the
        selection with a target.
        """

        self.removePropTargetPath(prop.GetPath(), target.GetPath())

    def getFocusProp(self):
        """Get the focus property from the property selection."""

        focusPath = self.getFocusPropPath()
        if focusPath is None:
            return None

        return self._getPropFromPath(focusPath)

    def getProps(self):
        """Get a list of all selected properties."""

        return [self._getPropFromPath(path)
            for path in self.getPropPaths()]

    def getPropTargets(self):
        """Get a dictionary which maps selected properties to a set of their
        selected targets or connections.
        """

        propTargets = OrderedDict()
        for propPath, targetPaths in self.getPropTargetPaths().items():

            prop = self._getPropFromPath(propPath)
            targets = {self._getTargetFromPath(target)
                for target in targetPaths}

            propTargets[prop] = targets

        return propTargets

    ### Computed Property Path Operations ###

    def clearComputedProps(self):
        """Clear the computed property selection."""

        self._computedPropSelection.clear()
        self._computedPropSelectionChanged()

    def addComputedPropPath(self, primPath, propName):
        """Add a computed property to the selection."""

        primPath = self._ensureValidPrimPath(primPath)
        self._validateComputedPropName(propName)

        self._computedPropSelection.addPropPath(primPath, propName)
        self._computedPropSelectionChanged()

    def removeComputedPropPath(self, primPath, propName):
        """Remove a computed property from the selection."""

        primPath = self._ensureValidPrimPath(primPath)
        self._validateComputedPropName(propName)

        self._computedPropSelection.removePropPath(primPath, propName)
        self._computedPropSelectionChanged()

    def setComputedPropPath(self, primPath, propName):
        """Clear the computed property selection, then add a single computed
        property path back to the selection.
        """

        primPath = self._ensureValidPrimPath(primPath)
        self._validateComputedPropName(propName)

        with self.batchComputedPropChanges:
            self.clearComputedProps()
            self.addComputedPropPath(primPath, propName)

    def getFocusComputedPropPath(self):
        """Get the focus computed property from the property selection."""

        self._requireNotBatchingComputedProps()

        propPaths = self._computedPropSelection.getPropPaths()
        if len(propPaths) > 0:
            return propPaths[-1]
        else:
            return (None, None)

    def getComputedPropPaths(self):
        """Get a list of all selected computed properties."""

        self._requireNotBatchingComputedProps()

        return self._computedPropSelection.getPropPaths()

    ### Computed Property Operations ###

    def addComputedProp(self, prop):
        """Add a computed property to the selection."""

        self.addComputedPropPath(prop.GetPrimPath(), prop.GetName())

    def removeComputedProp(self, prop):
        """Remove a computed property from the selection."""

        self.removeComputedPropPath(prop.GetPrimPath(), prop.GetName())

    def setComputedProp(self, prop):
        """Clear the computed property selection, then add a single computed
        property back to the selection.
        """

        self.setComputedPropPath(prop.GetPrimPath(), prop.GetName())

    def getFocusComputedProp(self):
        """Get the focus computed property from the property selection."""

        focusPath = self.getFocusComputedPropPath()
        if focusPath == (None, None):
            return None

        return self._getComputedPropFromPath(*focusPath)

    def getComputedProps(self):
        """Get a list of all selected computed properties."""

        return [self._getComputedPropFromPath(*path)
            for path in self.getComputedPropPaths()]



#
# Copyright 2017 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

# Functionality for expanding and collapsing legend components in usdview

LEGEND_BUTTON_SELECTEDSTYLE = ('background: rgb(189, 155, 84); '
                               'color: rgb(227, 227, 227);')

# Set the start/end points of an animation
def _SetAnimValues(anim, a1, a2):
    anim.setStartValue(a1)
    anim.setEndValue(a2)

# A function which takes a two-pane area and transforms it to
# open or close the bottom pane.
#
#    legendHeight
#    |   separator height
#    |   |  browser height
#    |   |  |->     ___________                ___________
#    |   |  |      |           |              |           |
#    |   |  |      |           |              |           |
#    |   |  |->    |           |    <--->     |           |
#    |---|-------> +++++++++++++              |           |
#    |             |           |    <--->     |           |
#    |             |           |              |           |
#    |----------->  -----------                +++++++++++
def ToggleLegendWithBrowser(legend, button, anim):
    legend.ToggleMinimized()

    # We are dragging downward, so collapse the legend and expand the
    # attribute viewer panel to take up the remaining space.
    if legend.IsMinimized():
        button.setStyleSheet('')
        _SetAnimValues(anim, legend.GetHeight(), 0)
    # We are expanding, so do the opposite.
    else:
        button.setStyleSheet(LEGEND_BUTTON_SELECTEDSTYLE)
        _SetAnimValues(anim, legend.GetHeight(), legend.GetResetHeight())

    anim.start()



#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

from .qt import QtGui, QtWidgets, QtCore
from pxr import Sdf
from .usdviewContextMenuItem import UsdviewContextMenuItem
from .common import (PropertyViewIndex, PropertyViewDataRoles,
                     PrimNotFoundException, PropertyNotFoundException)

#
# Specialized context menu for running commands in the attribute viewer.
#
class AttributeViewContextMenu(QtWidgets.QMenu):

    def __init__(self, parent, item, dataModel):
        QtWidgets.QMenu.__init__(self, parent)
        self._menuItems = _GetContextMenuItems(item, dataModel)

        for menuItem in self._menuItems:
            # create menu actions
            if menuItem.isValid() and menuItem.ShouldDisplay():
                action = self.addAction(menuItem.GetText(), menuItem.RunCommand)

                # set enabled
                if not menuItem.IsEnabled():
                    action.setEnabled(False)


def _GetContextMenuItems(item, dataModel):
    return [ # Root selection methods
             CopyAttributeNameMenuItem(dataModel, item),
             CopyAttributeValueMenuItem(dataModel, item),
             CopyAllTargetPathsMenuItem(dataModel, item),
             SelectAllTargetPathsMenuItem(dataModel, item),

             # Individual/multi target selection menus
             CopyTargetPathMenuItem(dataModel, item),
             SelectTargetPathMenuItem(dataModel, item),
             ViewSplineMenuItem(dataModel, item),]

def _selectPrimsAndProps(dataModel, paths):
    prims = []
    props = []
    for path in paths:

        primPath = path.GetAbsoluteRootOrPrimPath()
        prim = dataModel.stage.GetPrimAtPath(primPath)
        if not prim:
            raise PrimNotFoundException(primPath)
        prims.append(prim)

        if path.IsPropertyPath():
            prop = prim.GetProperty(path.name)
            if not prop:
                raise PropertyNotFoundException(path)
            props.append(prop)

    with dataModel.selection.batchPrimChanges:
        dataModel.selection.clearPrims()
        for prim in prims:
            dataModel.selection.addPrim(prim)

    with dataModel.selection.batchPropChanges:
        dataModel.selection.clearProps()
        for prop in props:
            dataModel.selection.addProp(prop)

    dataModel.selection.clearComputedProps()

#
# The base class for propertyview context menu items.
#
class AttributeViewContextMenuItem(UsdviewContextMenuItem):
    def __init__(self, dataModel, item):
        self._dataModel = dataModel
        self._item = item
        self._role = self._item.data(PropertyViewIndex.TYPE, QtCore.Qt.ItemDataRole.WhatsThisRole)
        self._name = self._item.text(PropertyViewIndex.NAME) if self._item else ""
        self._value = self._item.text(PropertyViewIndex.VALUE) if self._item else ""

    def IsEnabled(self):
        return True

    def ShouldDisplay(self):
        return True

    def GetText(self):
        return ""

    def RunCommand(self):
        pass

#
# Copy the attribute's name to clipboard.
#
class CopyAttributeNameMenuItem(AttributeViewContextMenuItem):
    def ShouldDisplay(self):
        return self._role not in (PropertyViewDataRoles.TARGET, PropertyViewDataRoles.CONNECTION)

    def GetText(self):
        return "Copy Property Name"

    def RunCommand(self):
        if self._name == "":
            return

        cb = QtWidgets.QApplication.clipboard()
        cb.setText(self._name, QtGui.QClipboard.Mode.Selection)
        cb.setText(self._name, QtGui.QClipboard.Mode.Clipboard)

#
# Copy the attribute's value to clipboard.
#
class CopyAttributeValueMenuItem(AttributeViewContextMenuItem):
    def ShouldDisplay(self):
        return self._role not in (PropertyViewDataRoles.TARGET, PropertyViewDataRoles.CONNECTION)

    def GetText(self):
        return "Copy Property Value"

    def RunCommand(self):
        # We display relationships targets as:
        #    /f, /g/a ...
        # But when we ask to copy the value, we'd like to get back:
        #    [Sdf.Path('/f'), Sdf.Path('/g/a')]
        # Which is useful for pasting into a python interpreter.
        if self._role == PropertyViewDataRoles.RELATIONSHIP_WITH_TARGETS:
            valueStr = str([Sdf.Path("".join(p.split())) \
                          for p in self._value.split(",")])
        else:
            valueStr = self._value
            if self._item:
                rawVal = self._item.rawValue
                if rawVal is not None:
                    valueStr = str(rawVal)

        cb = QtWidgets.QApplication.clipboard()
        cb.setText(valueStr, QtGui.QClipboard.Mode.Selection)
        cb.setText(valueStr, QtGui.QClipboard.Mode.Clipboard)

# --------------------------------------------------------------------
# Individual target selection menus
# --------------------------------------------------------------------

#
# Copy the target path to clipboard.
#
class CopyTargetPathMenuItem(AttributeViewContextMenuItem):
    def ShouldDisplay(self):
        return self._role in (PropertyViewDataRoles.TARGET, PropertyViewDataRoles.CONNECTION)

    def GetText(self):
        return "Copy Target Path As Text"

    def GetSelectedOfType(self):
        getRole = lambda s: s.data(PropertyViewIndex.TYPE, QtCore.Qt.ItemDataRole.WhatsThisRole)
        return [s for s in self._item.treeWidget().selectedItems() \
                    if getRole(s) in (PropertyViewDataRoles.TARGET, PropertyViewDataRoles.CONNECTION)]

    def RunCommand(self):
        if not self._item:
            return

        value = ", ".join([s.text(PropertyViewIndex.NAME) for s in self.GetSelectedOfType()])
        cb = QtWidgets.QApplication.clipboard()
        cb.setText(value, QtGui.QClipboard.Mode.Selection)
        cb.setText(value, QtGui.QClipboard.Mode.Clipboard)

#
# Jump to the target path in the prim browser
# This will include all other highlighted paths of this type, if any.
#
class SelectTargetPathMenuItem(CopyTargetPathMenuItem):
    def GetText(self):
        return "Select Target Path"

    def RunCommand(self):
        paths = [Sdf.Path(s.text(PropertyViewIndex.NAME))
            for s in self.GetSelectedOfType()]

        _selectPrimsAndProps(self._dataModel, paths)

# --------------------------------------------------------------------
# Target owning property selection menus
# --------------------------------------------------------------------

def _GetTargetPathsForItem(item):
    paths = [Sdf.Path(item.child(i).text(PropertyViewIndex.NAME)) for 
                 i in range (0, item.childCount())]
    # If there are no children, the value column must hold a valid path
    # (since the command is enabled).
    if len(paths) == 0:
        itemText = item.text(PropertyViewIndex.VALUE)
        if len(itemText) > 0:
            paths.append(Sdf.Path(itemText))
    return paths

#
# Select all target paths under the selected attribute
#
class SelectAllTargetPathsMenuItem(AttributeViewContextMenuItem):
    def ShouldDisplay(self):
        return (self._role == PropertyViewDataRoles.RELATIONSHIP_WITH_TARGETS
                or self._role == PropertyViewDataRoles.ATTRIBUTE_WITH_CONNNECTIONS)

    def IsEnabled(self):
        if not self._item:
            return False
        
        # Enable the menu item if there are one or more targets for this 
        # rel/attribute connection
        if self._item.childCount() > 0:
            return True

        # Enable the menu if the item holds a valid SdfPath.
        itemText = self._item.text(2)
        return Sdf.Path.IsValidPathString(itemText)

    def GetText(self):
        return "Select Target Path(s)"

    def RunCommand(self):
        if not self._item:
            return

        _selectPrimsAndProps(self._dataModel, 
                _GetTargetPathsForItem(self._item))

#
# Copy all target paths under the currently selected relationship to the clipboard
#
class CopyAllTargetPathsMenuItem(SelectAllTargetPathsMenuItem):
    def GetText(self):
        return "Copy Target Path(s) As Text"

    def RunCommand(self):
        if not self._item:
            return

        value = ", ".join(_GetTargetPathsForItem(self._item))
        
        cb = QtWidgets.QApplication.clipboard()
        cb.setText(value, QtGui.QClipboard.Mode.Selection)
        cb.setText(value, QtGui.QClipboard.Mode.Clipboard)

class ViewSplineMenuItem(AttributeViewContextMenuItem):
    _viewerByAttrPath = {}

    def ShouldDisplay(self):
        # Only show for attributes (not connections or targets)
        return self._role == PropertyViewDataRoles.ATTRIBUTE

    def IsEnabled(self):
        # Check if this attributes has a spline value source
        attr = self._getUsdAttribute()
        return attr and attr.HasSpline()

    def GetText(self):
        return "View Spline"

    def RunCommand(self):
        from .splineViewer import SplineViewer
        attr = self._getUsdAttribute()
        if not attr:
            return

        # Activate existing spline viewer if one exists for this spline values
        # attribute
        pathStr = str(attr.GetPath())
        if viewer := self._viewerByAttrPath.get(pathStr):
            viewer.raise_()
            viewer.activateWindow()
            return

        # Show a minimal dialog box with info
        splineViewer = SplineViewer(
            attr, currentFrame = self._dataModel.currentFrame)
        splineViewer.setAttribute(QtCore.Qt.WidgetAttribute.WA_DeleteOnClose)
        splineViewer.SetStartAndEndTime(
            self._dataModel.frameRangeBegin,
            self._dataModel.frameRangeEnd)
        # Store the viewer in a dictionary to keep it alive
        # until the user closes it
        self._viewerByAttrPath[pathStr] = splineViewer
        self._dataModel.currentFrameChanged.connect(
            splineViewer.SetCurrentFrame)
        self._dataModel.frameRangeChanged.connect(
            splineViewer.SetStartAndEndTime)
        # Remove the viewer from the dictionary when it is closed
        splineViewer.destroyed.connect(
            lambda: self._viewerByAttrPath.pop(pathStr, None))
        splineViewer.show()


    def _getUsdAttribute(self):
        stage = self._dataModel.stage
        if len(self._dataModel.selection.getPropPaths()) != 1:
            return None
        path = self._dataModel.selection.getPropPaths()[0]
        return stage.GetAttributeAtPath(path)



#
# Copyright 2019 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

from pxr import Tf
Tf.PreparePythonModule()
del Tf

from . import cameraArgs
from . import colorArgs
from . import complexityArgs
from . import framesArgs
from . import rendererArgs



#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
# Qt Components
from .qt import QtCore, QtGui, QtWidgets
from .common import KeyboardShortcuts

class AppEventFilter(QtCore.QObject):
    '''This class's primary responsibility is delivering key events to
    "the right place".  Given usdview's simplistic approach to shortcuts
    (i.e. just uses the native Qt mechanism that does not allow for 
    context-sensitive keypress dispatching), we take a simplistic approach
    to routing: use Qt's preferred mechanism of processing keyPresses 
    only in widgets that have focus; therefore, the primary behavior of this
    filter is to track mouse-position in order to set widget focus, so that
    widgets with keyboard navigation behaviors operate when the mouse is over
    them.
    
    We add one special behaviors on top of that, which is to turn unaccepted
    left/right events into up/down events for TreeView widgets, because we
    do not have a specialized class on which to provide this nice navigation
    behavior.'''

    # in future it would be a hotkey dispatcher instead of appController
    # that we'd dispatch to, but we don't have one yet
    def __init__(self, appController):
        QtCore.QObject.__init__(self)
        self._appController = appController
        
    def IsNavKey(self, key, modifiers):
        # Note that the arrow keys are considered part of the keypad on macOS.
        return (key in (QtCore.Qt.Key.Key_Left, QtCore.Qt.Key.Key_Right,
                        QtCore.Qt.Key.Key_Up, QtCore.Qt.Key.Key_Down,
                        QtCore.Qt.Key.Key_PageUp, QtCore.Qt.Key.Key_PageDown,
                        QtCore.Qt.Key.Key_Home, QtCore.Qt.Key.Key_End, 
                        KeyboardShortcuts.FramingKey)
                and modifiers in (QtCore.Qt.KeyboardModifier.NoModifier,
                                  QtCore.Qt.KeyboardModifier.KeypadModifier))
        
    def _IsWindow(self, obj):
        if isinstance(obj, QtWidgets.QWidget):
            return obj.isWindow()
        else:
            return isinstance(obj, QtGui.QWindow)
            
    def TopLevelWindow(self, obj):
        parent = obj.parent()
        return obj if (self._IsWindow(obj) or not parent) else self.TopLevelWindow(parent)

    def WantsNavKeys(self, w):
        if not w or self._IsWindow(w):
            return False
        # The broader test would be QtWidgets.QAbstractItemView,
        # but pragmatically, the TableViews in usdview don't really
        # benefit much from keyboard navigation, and we'd rather
        # allow the arrow keys drive the playhead when such widgets would
        # otherwise get focus
        elif isinstance(w, QtWidgets.QTreeView):
            return True
        else:
            return self.WantsNavKeys(w.parent())

    def NavigableOrTopLevelObject(self, w):
        if (not w or 
            self._IsWindow(w) or 
            isinstance(w, QtWidgets.QTreeView) or
            isinstance(w, QtWidgets.QDialog)):
            return w
        else:
            parent = w.parent()
            return w if not parent else self.NavigableOrTopLevelObject(parent)
        
    def JealousFocus(self, w):
        return (isinstance(w, QtWidgets.QLineEdit) or 
                isinstance(w, QtWidgets.QComboBox) or
                isinstance(w, QtWidgets.QTextEdit) or
                isinstance(w, QtWidgets.QPlainTextEdit) or
                isinstance(w, QtWidgets.QAbstractSlider) or
                isinstance(w, QtWidgets.QAbstractSpinBox) or
                isinstance(w, QtWidgets.QWidget) and w.windowModality() in [QtCore.Qt.WindowModality.WindowModal,
                                                                            QtCore.Qt.WindowModality.ApplicationModal])
            
    def SetFocusFromMousePos(self, backupWidget):
        # It's possible the mouse isn't over any of our windows at the time,
        # in which case use the top-level window of backupWidget.
        overObject = QtWidgets.QApplication.widgetAt(QtGui.QCursor.pos())
        topLevelObject = self.NavigableOrTopLevelObject(overObject)
        focusObject = topLevelObject if topLevelObject else self.TopLevelWindow(backupWidget)

        if focusObject and isinstance(focusObject, QtWidgets.QWidget):
            focusObject.setFocus()

    def eventFilter(self, widget, event):
        # There is currently no filtering we want to do for modal or popups
        if (QtWidgets.QApplication.activeModalWidget() or
            QtWidgets.QApplication.activePopupWidget()):
            return False
        
        currFocusWidget = QtWidgets.QApplication.focusWidget()

        # Check for ShortcutOverride events to ensure we pick up navigation keys
        # that have been set as shortcuts for QActions. We still want to 
        # dispatch those to the focus widget as needed.
        if (event.type() == QtCore.QEvent.Type.ShortcutOverride):
            if (self.IsNavKey(event.key(), event.modifiers()) and 
                    self.WantsNavKeys(currFocusWidget)):
                event.setAccepted(True)
                return True

        elif (event.type() == QtCore.QEvent.Type.KeyPress):
            key = event.key()

            isNavKey = self.IsNavKey(key, event.modifiers())
            if key == QtCore.Qt.Key.Key_Escape:
                # ESC resets focus based on mouse position, regardless of
                # who currently holds focus
                self.SetFocusFromMousePos(widget)
                return True
            elif currFocusWidget and self.JealousFocus(currFocusWidget):
                # Don't touch if there's a greedy focus widget
                return False
            elif (isNavKey and self.WantsNavKeys(currFocusWidget)):
                # Special handling for navigation keys:
                # 1. When a "navigable" widget is focussed (a TreeView), 
                #    route arrow keys to the widget and consume them
                # 2. To make for snappier navigation, when the TreeView
                #    won't accept a left/right because an item is already
                #    opened or closed, turn it into an up/down event.  It
                #    WBN if this behavior could be part of the widgets 
                #    themselves, but currently, usdview does not specialize
                #    a class for its TreeView widgets.
                event.setAccepted(False)
                currFocusWidget.event(event)
                accepted = event.isAccepted()
                if (not accepted  and 
                    key in (QtCore.Qt.Key.Key_Left, QtCore.Qt.Key.Key_Right)):
                    advance = (key == QtCore.Qt.Key.Key_Right)
                    altNavKey = QtCore.Qt.Key.Key_Down if advance else QtCore.Qt.Key.Key_Up
                    subEvent = QtGui.QKeyEvent(QtCore.QEvent.Type.KeyPress,
                                               altNavKey,
                                               event.modifiers())
                    QtWidgets.QApplication.postEvent(currFocusWidget, subEvent)
                event.setAccepted(True)
                return True
            elif isNavKey:
                if self._appController.processNavKeyEvent(event):
                    return True

        elif (event.type() == QtCore.QEvent.Type.MouseMove and 
              not self.JealousFocus(currFocusWidget)):
            self.SetFocusFromMousePos(widget)
            # Note we do not consume the event!
            
        # During startup, Qt seems to queue up events on objects that may
        # have disappeared by the time the eventFilter is called upon.  This
        # is true regardless of how late we install the eventFilter, and
        # whether we process pending events before installing.  So we 
        # silently ignore Runtime errors that occur as a result.
        try:
            return QtCore.QObject.eventFilter(self, widget, event)
        except RuntimeError:
            return True



#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
'''
Hopefully we can deprecate this since most of the array stuff is handled by the
arrayAttributeView
'''
from .qt import QtWidgets

def progressDialog(title, value):
    dialog = QtWidgets.QProgressDialog(title, "Cancel", 0, value)
    dialog.setModal(True)
    dialog.setMinimumDuration(500)
    return dialog

def prettyPrint(v):
    """Returns a string representing a "detailed view" of the value v.
    This string is used in the watch window"""

    # Pretty-print a dictionary
    if isinstance(v, dict):
        result = "Dictionary contents:\n"
        for pair in v.items():
            keystring = str(pair[0])
            valstring = prettyPrint(pair[1])
            result += "------\n%s:\n%s\n" % (keystring, valstring)

    # Pretty-print list
    elif isinstance(v, list):
        dialog = progressDialog("Pretty-printing list...", len(v))

        result = "[\n"
        for i in range(len(v)):
            dialog.setValue(i)
            result += str(i) + ": " + prettyPrint(v[i]) + "\n"
            if (dialog.wasCanceled()):
                return "Pretty-printing canceled"
        result += "]\n"
        dialog.done(0)

    # Pretty-print tuple
    elif isinstance(v, tuple):
        dialog = progressDialog("Pretty-printing tuple...", len(v))

        result = "(\n"
        for i in range(len(v)):
            dialog.setValue(i)
            result += str(i) + ": " + prettyPrint(v[i]) + "\n"
            if (dialog.wasCanceled()):
                return "Pretty-printing canceled"
        result += ")\n"
        dialog.done(0)
    else:
        from .scalarTypes import ToString
        result = ToString(v)

    return result




#
# Copyright 2019 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

import argparse

def GetAllPluginArguments():
    """
    Returns argument strings for all the renderer plugins available.
    """

    from pxr import UsdImagingGL
    return [ UsdImagingGL.Engine.GetRendererDisplayName(pluginId) for
                pluginId in UsdImagingGL.Engine.GetRendererPlugins() ]


def GetPluginIdFromArgument(argumentString):
    """
    Returns plugin id, if found, for the passed in argument string.

    Valid argument strings are returned by GetAllPluginArguments().
    """

    from pxr import UsdImagingGL
    for p in UsdImagingGL.Engine.GetRendererPlugins():
        if argumentString == UsdImagingGL.Engine.GetRendererDisplayName(p) or argumentString == p:
            return p
    return None


def AddCmdlineArgs(argsParser, altHelpText=''):
    """
    Adds Hydra renderer-related command line arguments to argsParser.

    The resulting 'rendererPlugin' argument will be a _RendererPlugin instance
    representing one of the available Hydra renderer plugins.
    """
    from pxr import UsdImagingGL

    helpText = altHelpText
    if not helpText:
        helpText = (
            'Hydra renderer plugin to use when generating images. "GL" and '
            '"Storm" currently alias to the same renderer, Storm.')

    class RendererChoices(list):
        def __contains__(self, other):
            return super().__contains__(other) or UsdImagingGL.Engine.GetRendererPlugins().__contains__(other)

    renderers = RendererChoices(GetAllPluginArguments())
    # "GL" is still (unfortunately) the offical display name for Storm, but 
    # we've hacked UsdImagingGLEngine::GetRendererDisplayName to instead 
    # return "Storm" for the HdStormRendererPlugin. We still wish to support 
    # "GL" as an argument here, however. Both "GL" and "Storm" refer to the 
    # Storm renderer.
    renderers.append('GL')

    class ResolveStormAlias(argparse.Action):
        def __call__(self, parser, namespace, value, option_string):
            if value == 'Storm':
                setattr(namespace, self.dest, 'GL')
            else:
                setattr(namespace, self.dest, value)

    argsParser.add_argument('--renderer', '-r',
        dest='rendererPlugin',
        choices=renderers,
        help=helpText,
        action=ResolveStormAlias)
    



#
# Copyright 2021 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

from pxr import Tf, Sdf, Sdr, Usd, UsdShade, UsdUI, UsdUtils, Vt
from pxr.UsdUtils.constantsGroup import ConstantsGroup

class SchemaDefiningKeys(ConstantsGroup):
    API_SCHEMAS_FOR_ATTR_PRUNING = "apiSchemasForAttrPruning"
    API_SCHEMA_AUTO_APPLY_TO = "apiSchemaAutoApplyTo"
    API_SCHEMA_CAN_ONLY_APPLY_TO = "apiSchemaCanOnlyApplyTo"
    IS_USD_SHADE_CONTAINER = "isUsdShadeContainer"
    SCHEMA_PROPERTY_NS_PREFIX_OVERRIDE = "schemaPropertyNSPrefixOverride"
    PROVIDES_USD_SHADE_CONNECTABLE_API_BEHAVIOR = \
            "providesUsdShadeConnectableAPIBehavior"
    REQUIRES_USD_SHADE_ENCAPSULATION = "requiresUsdShadeEncapsulation"
    SCHEMA_BASE = "schemaBase"
    SCHEMA_KIND = "schemaKind"
    SCHEMA_NAME = "schemaName"
    TF_TYPENAME_SUFFIX = "tfTypeNameSuffix"
    TYPED_SCHEMA_FOR_ATTR_PRUNING = "typedSchemaForAttrPruning"

class SchemaDefiningMiscConstants(ConstantsGroup):
    API_SCHEMA_BASE = "APISchemaBase"
    API_STRING = "API"
    NodeDefAPI = "NodeDefAPI"
    SINGLE_APPLY_SCHEMA = "singleApply"
    TYPED_SCHEMA = "Typed"
    USD_SOURCE_TYPE = "USD"

class PropertyDefiningKeys(ConstantsGroup):
    CONNECTABILITY = "connectability"
    INTERNAL_DISPLAY_GROUP = "Internal"
    NULL_VALUE = "null"
    PROPERTY_NS_PREFIX_OVERRIDE = "propertyNSPrefixOverride"
    SDF_VARIABILITY_UNIFORM_STRING = "Uniform"
    SHADER_ID = "shaderId"
    USD_SUPPRESS_PROPERTY = "usdSuppressProperty"
    USD_VARIABILITY = "usdVariability"
    WIDGET = "widget"
    MIN = "min"
    MAX = "max"
    SLIDER_MIN = "slidermin"
    SLIDER_MAX = "slidermax"
    STRICT_LIMITS = "strictLimits"

class UserDocConstants(ConstantsGroup):
    USERDOC_FULL = "userDoc"
    USERDOC_BRIEF = "userDocBrief"

def _IsNSPrefixConnectableAPICompliant(nsPrefix):
    return (nsPrefix == UsdShade.Tokens.inputs[:1] or \
            nsPrefix == UsdShade.Tokens.outputs[:1])

def _CreateAttrSpecFromNodeAttribute(primSpec, prop, stage,
        primDefForAttrPruning, schemaPropertyNSPrefixOverride, isSdrInput=True):
    propMetadata = prop.GetMetadata()
    # Early out if the property should be suppressed from being translated to
    # propertySpec
    if ((PropertyDefiningKeys.USD_SUPPRESS_PROPERTY in propMetadata) and
            propMetadata[PropertyDefiningKeys.USD_SUPPRESS_PROPERTY] == "True"):
        return None

    propertyNSPrefixOverride = schemaPropertyNSPrefixOverride
    if PropertyDefiningKeys.PROPERTY_NS_PREFIX_OVERRIDE in propMetadata:
        propertyNSPrefixOverride = \
            propMetadata[PropertyDefiningKeys.PROPERTY_NS_PREFIX_OVERRIDE]

    propName = prop.GetName()

    # Error out if trying to use an explicit propertyNSPrefixOverride on an
    # output attr
    if (not isSdrInput and propertyNSPrefixOverride is not None and \
            propertyNSPrefixOverride != UsdShade.Tokens.outputs[:-1]):
        Tf.RaiseRuntimeError("Presence of (%s) output parameter contradicts " \
            "the presence of propertyNSPrefixOverride (\"%s\"), as it is " \
            "illegal for non-shader nodes to contain output parameters, or " \
            "shader nodes' outputs to not have the \"outputs\" namespace " \
            "prefix." %(propName, propertyNSPrefixOverride))

    attrType = prop.GetTypeAsSdfType().GetSdfType()
    
    if not Sdf.Path.IsValidNamespacedIdentifier(propName):
        Tf.RaiseRuntimeError("Property name (%s) for schema (%s) is an " \
                "invalid namespace identifier." %(propName, primSpec.name))

    # if propertyNSPrefixOverride is provided and we are an output then already
    # thrown exception
    # Note that UsdShade inputs and outputs tokens contain the ":" delimiter, so
    # we need to strip this to be used with JoinIdentifier
    if propertyNSPrefixOverride is None:
        propertyNSPrefixOverride = UsdShade.Tokens.inputs[:-1] if isSdrInput \
                else UsdShade.Tokens.outputs[:-1]

    # Apply propertyNSPrefixOverride
    propName = Sdf.Path.JoinIdentifier([propertyNSPrefixOverride, propName])

    # error and early out if duplicate property on primDefForAttrPruning exists
    # and has different types
    if primDefForAttrPruning:
        primDefAttr = primDefForAttrPruning.GetAttributeDefinition(propName)
        if primDefAttr:
            usdAttrType = primDefAttr.GetTypeName()
            if (usdAttrType != attrType):
                Tf.Warn("Generated schema's property type '%s', "
                        "differs usd schema's property type '%s', for "
                        "duplicated property '%s'" %(attrType, usdAttrType, 
                        propName))
            return propName

    # Copy over property parameters
    options = prop.GetOptions()
    if options and attrType == Sdf.ValueTypeNames.String:
        attrType = Sdf.ValueTypeNames.Token

    attrVariability = Sdf.VariabilityUniform \
            if ((PropertyDefiningKeys.USD_VARIABILITY in propMetadata) and
                propMetadata[PropertyDefiningKeys.USD_VARIABILITY] == 
                    PropertyDefiningKeys.SDF_VARIABILITY_UNIFORM_STRING) \
                            else Sdf.VariabilityVarying
    attrSpec = Sdf.AttributeSpec(primSpec, propName, attrType,
            attrVariability)

    # Look up attrSpec's UsdAttribute so we can use UsdUI to populate
    # hint values
    attr = stage.GetAttributeAtPath(attrSpec.path)
    if not attr:
        Tf.RaiseCodingError(
            "Failed to find composed attribute (%s) on " \
            "schema (%s)" % (propName, primSpec.name))

    attrHints = UsdUI.AttributeHints(attr)
    if not attrHints:
        Tf.RaiseCodingError(
            "Failed to populate UI hints for attribute (%s) on " \
            "schema (%s)" % (propName, primSpec.name))

    if PropertyDefiningKeys.WIDGET in prop.GetMetadata().keys():
        if (prop.GetMetadata()[PropertyDefiningKeys.WIDGET] == \
          PropertyDefiningKeys.NULL_VALUE):
            attrHints.SetHidden(True)

    if prop.GetHelp():
        _SetSchemaUserDocFields(attrSpec, prop.GetHelp())
    elif prop.GetLabel(): # fallback documentation can be label
        _SetSchemaUserDocFields(attrSpec, prop.GetLabel())

    if prop.GetPage():
        attrHints.SetDisplayGroup(prop.GetPage())
    if prop.GetLabel():
        attrHints.SetDisplayName(prop.GetLabel())
    if prop.GetShownIf():
        attrHints.SetShownIf(prop.GetShownIf())

    _PopulateArraySizeConstraint(attr, prop)
    _PopulateLimits(attr, prop)
    _PopulateOptions(attrHints, attrSpec, prop, propName, attrType)

    attrSpec.default = prop.GetDefaultValueAsSdfType()

    # The input property should remain connectable (interfaceOnly)
    # even if sdrProperty marks the input as not connectable
    if propertyNSPrefixOverride == UsdShade.Tokens.inputs[:-1] and \
            not prop.IsConnectable():
        attrSpec.SetInfo(PropertyDefiningKeys.CONNECTABILITY, 
                UsdShade.Tokens.interfaceOnly)

    return propName

def _SetSchemaUserDocFields(spec, doc):
    """
    Sets the user doc custom metadata fields in the generated schema for prim 
    and attribute specs. 
    """
    # Set the "brief" user doc, used for in-context help, e.g. in DCC tools.
    # We currently want the full content, so we don't shorten userDocBrief.
    spec.customData[UserDocConstants.USERDOC_BRIEF] = doc
    # Set the "long-form" user doc, used when generating HTML schema docs
    # (example: https://openusd.org/release/user_guides/schemas/index.html)
    spec.customData[UserDocConstants.USERDOC_FULL] = doc


def StringToBool(val):
    """Convert a string representation of truth to True or False.
    
    True values are 'y', 'yes', 't', 'true', 'on', and '1';
    False values are 'n', 'no', 'f', 'false', 'off', and '0'.
    
    Raises ValueError if `val` is anything else.
    """
    val = val.lower()
    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return True
    elif val in ('n', 'no', 'f', 'false', 'off', '0'):
        return False
    else:
        raise ValueError(f"Invalid truth value: {val}")

def _PopulateArraySizeConstraint(attr, sdrProp):
    if not sdrProp.IsArray():
        return

    # Sdr int and float arrays of size 2, 3, and 4 get turned into statically
    # shaped GfVec "tuples" in USD (e.g., int[3] in a shader becomes int3 in
    # USD). Since element count is built into these types, we never want to
    # write arraySizeConstraint for them.
    staticSizeTypes = [
        Sdf.ValueTypeNames.Int2,
        Sdf.ValueTypeNames.Int3,
        Sdf.ValueTypeNames.Int4,
        Sdf.ValueTypeNames.Float2,
        Sdf.ValueTypeNames.Float3,
        Sdf.ValueTypeNames.Float4,
    ]
    if sdrProp.GetTypeAsSdfType().GetSdfType() in staticSizeTypes:
        return

    # Determine whether a fixed arraySize is present. We want to avoid writing
    # arraySizeConstraint=0 since that's the fallback value.
    arraySize = sdrProp.GetArraySize()
    hasFixedArraySize = arraySize > 0 and not sdrProp.IsDynamicArray()

    tupleSize = sdrProp.GetTupleSize()

    # Translate arraySize and tupleSize values into the USD encoding:
    #   - arraySizeConstraint == 0 indicates the array is dynamic and its size
    #     is unrestricted
    #   - arraySizeConstraint > 0 indicates the exact, fixed size of the array
    #   - arraySizeConstraint < 0 indicates the array is dynamic, but
    #     `N=abs(arraySizeConstraint)` is the tuple-length
    if tupleSize > 0:
        attr.SetArraySizeConstraint(-tupleSize)

        # arraySizeConstraint is intentionally a single value to prevent
        # encoding inconsistent arraySize and tupleSize values. This means we
        # can't write both values, even if they *are* consistent with each
        # other (which would correspond to a fixed number of tuples).
        #
        # If both arraySize and tupleSize were specified, issue a warning, and
        # point out if they are inconsistent with each other.
        if hasFixedArraySize:
            consistentStr = 'inconsistent ' \
                if arraySize % tupleSize != 0 else ''
            Tf.Warn("Ignoring %sarraySize (%d) specified with tupleSize (%d) "
                    "for attribute (%s) on schema (%s)" % \
                    (consistentStr, arraySize, tupleSize, attr.GetName(),
                     attr.GetPrim().GetName()))
    elif hasFixedArraySize:
        attr.SetArraySizeConstraint(arraySize)

def _ParseLimitsValue(key, attr, sdrProp):
    # Parse and return the value at the given key in metadata, assuming the
    # same value type as sdrProp
    if v := sdrProp.GetMetadata().get(key):
        parsed, err = Sdr.MetadataHelpers.ParseSdfValue(v, sdrProp)
        if not err:
            return parsed
        else:
            Tf.Warn("Failed to parse %s value ('%s') for attribute %s " \
                    "on schema %s: %s" % \
                    (key, v, attr.GetName(), attr.GetPrim().GetName(), err))
    return None

def _SetLimits(limits, minimum, maximum):
    if minimum is not None:
        limits.SetMinimum(minimum)
    if maximum is not None:
        limits.SetMaximum(maximum)

def _PopulateLimits(attr, sdrProp):
    # Parse raw limits values out of metadata
    #
    # XXX In the future, SdrShaderProperty should provide a clean API for
    # accessing limits values that hides these parsing details and all the
    # hard vs soft logic below
    minimum = _ParseLimitsValue(PropertyDefiningKeys.MIN, attr, sdrProp)
    maximum = _ParseLimitsValue(PropertyDefiningKeys.MAX, attr, sdrProp)
    sliderMin = _ParseLimitsValue(PropertyDefiningKeys.SLIDER_MIN, attr, sdrProp)
    sliderMax = _ParseLimitsValue(PropertyDefiningKeys.SLIDER_MAX, attr, sdrProp)

    strict = None
    if v := sdrProp.GetMetadata().get(PropertyDefiningKeys.STRICT_LIMITS):
        try:
            strict = StringToBool(v)
        except:
            Tf.Warn("Failed to parse strictLimits value ('%s') for attribute " \
                    "%s on schema %s" % \
                    (v, attr.GetName(), attr.GetPrim().GetName()))

    # Now determine which values correspond to hard and soft limits.
    #
    # In general, the "regular" min and max keys become hard limits, and
    # slidermin/slidermax become soft limits. The strictLimits key can override
    # this behavior to make one or the other change destinations:
    #
    #  * strictLimits=True will cause slidermin/slidermax to become hard limits,
    #    but only if neither min nor max are set
    #  * strictLimits=False will cause min/max to become soft limits, regardless
    #    of whether slidermin/slidermax are set
    #
    # Values can be sparsely specified, e.g., slidermin can be specified
    # without a corresponding slidermax.
    sliderValsSpecified = sliderMin is not None or sliderMax is not None
    regularValsSpecified = minimum is not None or maximum is not None

    if sliderValsSpecified:
        if regularValsSpecified:
            if strict in (None, True):
                # No override, apply both sets of values as-is
                _SetLimits(attr.GetSoftLimits(), sliderMin, sliderMax)
                _SetLimits(attr.GetHardLimits(), minimum, maximum)
            else:
                # strictLimits explicitly false, clobber sliderMin/sliderMax
                # with min/max as soft limits
                _SetLimits(attr.GetSoftLimits(), minimum, maximum)
        elif strict in (None, False):
            # No override, sliderMin/sliderMax become soft limits
            _SetLimits(attr.GetSoftLimits(), sliderMin, sliderMax)
        else:
            # strictLimits explicitly true, promote sliderMin/sliderMax into
            # hard limits
            _SetLimits(attr.GetHardLimits(), sliderMin, sliderMax)
    elif regularValsSpecified:
        if strict in (None, True):
            # No override, min/max become hard limits
            _SetLimits(attr.GetHardLimits(), minimum, maximum)
        else:
            # strictLimits explicitly false, demote min/max into soft limits
            _SetLimits(attr.GetSoftLimits(), minimum, maximum)

def _ParseOptionValues(values, attrSpec, sdrProp, propName):
    result = []
    for v in values:
        parsedVal, err = Sdr.MetadataHelpers.ParseSdfValue(v, sdrProp)

        if not err:
            result.append(parsedVal)
        else:
            Tf.Warn("Failed to parse option value ('%s') for attribute (%s) " \
                    "on schema (%s): %s" % \
                    (v, propName, attrSpec.owner.name, err))
            return []

    return result

def _PopulateOptions(attrHints, attrSpec, sdrProp, propName, attrType):
    # Translate sdrProp's `options` field into either allowedTokens or
    # valueLabels
    #
    # Neither apply to asset params, which use the options hint to indicate
    # when a texture asset is expected.
    options = sdrProp.GetOptions()
    if not options or sdrProp.IsAssetIdentifier():
        return

    # `options` is a list of (name, value) tuples that represents either
    # allowedTokens or valueLabels. If both names and values are provided,
    # encode as valueLabels. If only one side is provided (and the attribute is
    # token-valued), author allowedTokens.
    #
    # To continue generating backwards-compatible schemas on a temporary basis,
    # if USD_POPULATE_LEGACY_ALLOWED_TOKENS is set we always populate
    # allowedTokens even if also populating valueLabels.
    names, vals = zip(*options)
    namesProvided = not all(len(n) == 0 for n in names)
    valsProvided = not all(len(v) == 0 for v in vals)

    isTokenValued = attrType == Sdf.ValueTypeNames.Token

    populateLabels = namesProvided and valsProvided
    populateTokens = isTokenValued and \
        (not populateLabels or
         Tf.GetEnvSetting('USD_POPULATE_LEGACY_ALLOWED_TOKENS'))

    if populateLabels:
        # Write labels, using values that have been parsed to the appropriate
        # value type for the attribute (the elements of `vals` are all strings)
        parsedVals = _ParseOptionValues(vals, attrSpec, sdrProp, propName)
        if parsedVals:
            attrHints.SetValueLabels(
                dict(zip(names, parsedVals)))
            attrHints.SetValueLabelsOrder(names)
        else:
            Tf.Warn("Could not set valueLabels for attribute (%s) on " \
                    "schema (%s)" % (propName, attrSpec.owner.name))

    if populateTokens:
        # Write allowedTokens. Usually only names are provided in this case,
        # but prefer the values if available.
        if valsProvided:
            attrSpec.allowedTokens = vals
        elif namesProvided:
            attrSpec.allowedTokens = names

    # Complain if allowed tokens were provided for a non-token-valued param
    if not isTokenValued and not populateLabels:
        Tf.Warn("Ignoring allowedTokens provided for non-token (%s) " \
                "valued attribute (%s) on schema (%s): %s" \
                % (attrType, propName, attrSpec.owner.name, options))

def UpdateSchemaWithSdrNode(schemaLayer, sdrNode, renderContext="",
        overrideIdentifier=""):
    """
    Updates the given schemaLayer with primSpec and propertySpecs from sdrNode
    metadata. 

    A renderContext can be provided which is used in determining the
    shaderId namespace, which follows the pattern: 
    "<renderContext>:<SdrShaderNodeContext>:shaderId". Note that we are using a
    node's context (SDR_NODE_CONTEXT_TOKENS) here to construct the shaderId
    namespace, so shader parsers should make sure to use appropriate
    SDR_NODE_CONTEXT_TOKENS in the node definitions.

    overrideIdentifier parameter is the identifier which should be used when 
    the identifier of the node being processed differs from the one Sdr will 
    discover at runtime, such as when this function is def a node constructed 
    from an explicit asset path. This should only be used when clients know the 
    identifier being passed is the true identifier which sdr Runtime will 
    provide when querying using GetShaderNodeByIdentifierAndType, etc.

    It consumes the following attributes (that manifest as Sdr 
    metadata) in addition to many of the standard Sdr metadata
    specified and parsed (via its parser plugin).

    Node Level Metadata:
        - "schemaName": Name of the new schema populated from the given sdrNode
          (Required)
        - "schemaKind": Specifies the UsdSchemaKind for the schema being
          populated from the sdrNode. (Note that this does not support
          multiple apply schema kinds).
        - "schemaBase": Base schema from which the new schema should inherit
          from. Note this defaults to "APISchemaBase" for an API schema or 
          "Typed" for a concrete scheme.
        - "apiSchemasForAttrPruning": A list of core API schemas which will be
          composed together and any shared shader property from this prim
          definition is pruned from the resultant schema. 
        - "typedSchemaForAttrPruning": A core typed schema which will be
          composed together with the apiSchemasForAttrPruning and any shared 
          shader property from this prim definition is pruned from the 
          resultant schema. If no typedSchemaForAttrPruning is provided then 
          only the apiSchemasForAttrPruning are composed to create a prim 
          definition. This will only be used when creating an APISchema.
        - "apiSchemaAutoApplyTo": The schemas to which the sdrNode populated 
          API schema will autoApply to.
        - "apiSchemaCanOnlyApplyTo": If specified, the API schema generated 
          from the sdrNode can only be validly applied to this set of schemas.
        - "providesUsdShadeConnectableAPIBehavior": Used to enable a 
          connectability behavior for an API schema.
        - "isUsdShadeContainer": Only used when
          providesUsdShadeConnectableAPIBehavior is set to true. Marks the
          connectable prim as a UsdShade container type.
        - "requiresUsdShadeEncapsulation": Only used when
          providesUsdShadeConnectableAPIBehavior is set to true. Configures the
          UsdShade encapsulation rules governing its connectableBehavior.
        - "tfTypeNameSuffix": Class name which will get registered with TfType 
          system. This gets appended to the domain name to register with TfType.
        - "schemaPropertyNSPrefixOverride": Node level metadata which can drive
          all node's properties namespace prefix. This can be useful for
          non connectable nodes which should not get UsdShade inputs and outputs
          namespace prefix.

    Property Level Metadata:
        - "usdVariability": Property level metadata which specifies a specific 
          sdrNodeProperty should have its USD variability set to Uniform or 
          Varying
        - "usdSuppressProperty": A property level metadata which determines if 
          the property should be suppressed from translation from args to 
          property spec.
        - "propertyNSPrefixOverride": Provides a way to override a property's
          namespace from the default (inputs:/outputs:) or from a node's
          schemaPropertyNSPrefixOverride metadata.

    Sdr Property Metadata to SdfPropertySpec Translations
        - A "null" value for Widget SdrShaderProperty metadata translates to
          'hidden' in the SdfPropertySpec's uiHints metadata dictionary.
        - SdrShaderProperty's Help metadata (Label metadata if Help metadata not
          provided) translates to the SdfPropertySpec's userDocBrief and userDoc
          custom metadata strings.  
        - SdrShaderProperty's Page metadata translates to 'displayGroup'
          in the SdfPropertySpec's uiHints metadata dictionary.
        - SdrShaderProperty's Label metadata translates to 'displayName'
          in the SdfPropertySpec's uiHints metadata dictionary.
        - SdrShaderProperty's Options translates to either 'allowedTokens'
          (when either labels or values are provided) or 'valueLabels' (when
          both labels and values are provided) in the SdfPropertySpec's uiHints
          metadata dictionary. For backwards compatibility, set
          USD_POPULATE_LEGACY_ALLOWED_TOKENS to write allowedTokens in addition
          to valueLabels for string- and token-valued attributes that provide
          both labels and values.
        - SdrShaderProperty's Default value translates to the SdfPropertySpec's
          Default value.
        - Connectable input properties translates to InterfaceOnly
          SdfPropertySpec's CONNECTABILITY.
    """

    import os

    # Early exit on invalid parameters
    if not schemaLayer:
        Tf.Warn("No Schema Layer provided")
        return
    if sdrNode is None:
        # This is a workaround to iterate through invalid sdrNodes (nodes not 
        # having any input or output properties). Currently these nodes return
        # false when queried for IsValid().
        # Refer: pxr/usd/ndr/node.h#140-149
        Tf.Warn("No valid sdrNode provided")
        return

    sdrNodeMetadata = sdrNode.GetMetadata()

    if SchemaDefiningKeys.SCHEMA_NAME not in sdrNodeMetadata:
        Tf.Warn("Sdr Node (%s) does not define a schema name metadata." \
                %(sdrNode.GetName()))
        return
    schemaName = sdrNodeMetadata[SchemaDefiningKeys.SCHEMA_NAME]
    if not Tf.IsValidIdentifier(schemaName):
        Tf.RaiseRuntimeError("schemaName (%s) is an invalid identifier; "
                "Provide a valid USD identifer for schemaName, example (%s) "
                %(schemaName, Tf.MakeValidIdentifier(schemaName)))

    tfTypeNameSuffix = None
    if SchemaDefiningKeys.TF_TYPENAME_SUFFIX in sdrNodeMetadata:
        tfTypeNameSuffix = sdrNodeMetadata[SchemaDefiningKeys.TF_TYPENAME_SUFFIX]
        if not Tf.IsValidIdentifier(tfTypeNameSuffix):
            Tf.RaiseRuntimeError("tfTypeNameSuffix (%s) is an invalid " \
                    "identifier" %(tfTypeNameSuffix))

    if SchemaDefiningKeys.SCHEMA_KIND not in sdrNodeMetadata:
        schemaKind = SchemaDefiningMiscConstants.TYPED_SCHEMA
    else:
        schemaKind = sdrNodeMetadata[SchemaDefiningKeys.SCHEMA_KIND]

    # Note: We are not working on dynamic multiple apply schemas right now.
    isAPI = schemaKind == SchemaDefiningMiscConstants.SINGLE_APPLY_SCHEMA
    isTyped = schemaKind == SchemaDefiningMiscConstants.TYPED_SCHEMA

    # Fix schemaName and warn if needed
    if isAPI and \
        not schemaName.endswith(SchemaDefiningMiscConstants.API_STRING):
        Tf.Warn("node metadata implies the generated schema being created is "
        "an API schema, fixing schemaName to reflect that")
        schemaName = schemaName + SchemaDefiningMiscConstants.API_STRING

    if isAPI and tfTypeNameSuffix and \
        not tfTypeNameSuffix.endswith(SchemaDefiningMiscConstants.API_STRING):
            Tf.Warn("node metadata implies the generated schema being created "
            "is an API schema, fixing tfTypeNameSuffix to reflect that")
            tfTypeNameSuffix = tfTypeNameSuffix + \
                    SchemaDefiningMiscConstants.API_STRING

    if SchemaDefiningKeys.SCHEMA_BASE not in sdrNodeMetadata:
        Tf.Warn("No schemaBase specified in node metadata, defaulting to "
                "APISchemaBase for API schemas else Typed")
        schemaBase = SchemaDefiningMiscConstants.API_SCHEMA_BASE if isAPI \
                else SchemaDefiningMiscConstants.TYPED_SCHEMA
    else:
        schemaBase = sdrNodeMetadata[SchemaDefiningKeys.SCHEMA_BASE]

    apiSchemaAutoApplyTo = None
    if SchemaDefiningKeys.API_SCHEMA_AUTO_APPLY_TO in sdrNodeMetadata:
        apiSchemaAutoApplyTo = \
            sdrNodeMetadata[SchemaDefiningKeys.API_SCHEMA_AUTO_APPLY_TO] \
                .split('|')

    apiSchemaCanOnlyApplyTo = None
    if SchemaDefiningKeys.API_SCHEMA_CAN_ONLY_APPLY_TO in sdrNodeMetadata:
        apiSchemaCanOnlyApplyTo = \
            sdrNodeMetadata[SchemaDefiningKeys.API_SCHEMA_CAN_ONLY_APPLY_TO] \
                .split('|')

    providesUsdShadeConnectableAPIBehavior = False
    if SchemaDefiningKeys.PROVIDES_USD_SHADE_CONNECTABLE_API_BEHAVIOR in \
            sdrNodeMetadata:
        providesUsdShadeConnectableAPIBehavior = \
            StringToBool(sdrNodeMetadata[SchemaDefiningKeys. \
                PROVIDES_USD_SHADE_CONNECTABLE_API_BEHAVIOR])

    apiSchemasForAttrPruning = None
    if SchemaDefiningKeys.API_SCHEMAS_FOR_ATTR_PRUNING in sdrNodeMetadata:
        apiSchemasForAttrPruning = \
            sdrNodeMetadata[SchemaDefiningKeys.API_SCHEMAS_FOR_ATTR_PRUNING] \
                .split('|')

    typedSchemaForAttrPruning = ""
    if isAPI and \
            SchemaDefiningKeys.TYPED_SCHEMA_FOR_ATTR_PRUNING in sdrNodeMetadata:
        typedSchemaForAttrPruning = \
            sdrNodeMetadata[SchemaDefiningKeys.TYPED_SCHEMA_FOR_ATTR_PRUNING]

    schemaPropertyNSPrefixOverride = None
    if SchemaDefiningKeys.SCHEMA_PROPERTY_NS_PREFIX_OVERRIDE in sdrNodeMetadata:
        schemaPropertyNSPrefixOverride = \
            sdrNodeMetadata[ \
                SchemaDefiningKeys.SCHEMA_PROPERTY_NS_PREFIX_OVERRIDE]

    usdSchemaReg = Usd.SchemaRegistry()

    # determine if the node being processed provides UsdShade-Connectability, 
    # this helps in determining what namespace to use and also to report error 
    # if a non-connectable node has outputs properties, which is malformed.
    # - Does the node derive from a schemaBase which provides connectable
    # behavior. Warn if schemaPropertyNSPrefixOverride is also specified, as 
    # these metadata won't be used.
    # - If no schemaBase then we default to UsdShade connectable node's 
    # inputs:/outputs: namespace prefix, unless schemaPropertyNSPrefixOverride 
    # is provided. 
    # - We also report an error if schemaPropertyNSPrefixOverride is provided 
    # and an output property is found on the node being processed.
    schemaBaseProvidesConnectability = UsdShade.ConnectableAPI. \
            HasConnectableAPI(usdSchemaReg.GetTypeFromName(schemaBase))

    emitSdrOutput = True
    for outputName in sdrNode.GetShaderOutputNames():
        if PropertyDefiningKeys.USD_SUPPRESS_PROPERTY in \
                sdrNode.GetShaderOutput(outputName).GetMetadata():
            emitSdrOutput = False
            break;

    if (emitSdrOutput and \
        len(sdrNode.GetShaderOutputNames()) > 0 and \
        schemaPropertyNSPrefixOverride is not None and \
        not _IsNSPrefixConnectableAPICompliant( \
                schemaPropertyNSPrefixOverride)):
        Tf.RaiseRuntimeError("Presence of (%s) output parameters contradicts " \
            "the presence of schemaPropertyNSPrefixOverride (\"%s\"), as it " \
            "is illegal for non-connectable nodes to contain output " \
            "parameters, or shader nodes' outputs to not have the \"outputs\"" \
            "namespace prefix." %(len(sdrNode.GetShaderOutputNames()), \
            schemaPropertyNSPrefixOverride))

    if (schemaBaseProvidesConnectability and \
            schemaPropertyNSPrefixOverride is not None and \
            not _IsNSPrefixConnectableAPICompliant( \
                schemaPropertyNSPrefixOverride)):
        Tf.Warn("Node %s provides UsdShade-Connectability as it derives from " \
                "%s, schemaPropertyNSPrefixOverride \"%s\" will not be used." \
                %(schemaName, schemaBase, schemaPropertyNSPrefixOverride))
        # set schemaPropertyNSPrefixOverride to "inputs", assuming default 
        # UsdShade Connectability namespace prefix
        schemaPropertyNSPrefixOverride = "inputs"

    primSpec = schemaLayer.GetPrimAtPath(schemaName)

    if (primSpec):
        # if primSpec already exist, remove entirely and recreate using the 
        # parsed sdr node
        if primSpec.nameParent:
            del primSpec.nameParent.nameChildren[primSpec.name]
        else:
            del primSpec.nameRoot.nameChildren[primSpec.name]

    primSpec = Sdf.PrimSpec(schemaLayer, schemaName, Sdf.SpecifierClass,
            "" if isAPI else schemaName)
    
    primSpec.inheritPathList.explicitItems = ["/" + schemaBase]

    primSpecCustomData = {}
    if isAPI:
        primSpecCustomData["apiSchemaType"] = schemaKind 
    if tfTypeNameSuffix:
        # Defines this classname for TfType system
        # can help avoid duplicate prefix with domain and className
        # Tf type system will automatically pick schemaName as tfTypeName if
        # this is not set!
        primSpecCustomData["className"] = tfTypeNameSuffix

    if apiSchemaAutoApplyTo:
        primSpecCustomData['apiSchemaAutoApplyTo'] = \
            Vt.TokenArray(apiSchemaAutoApplyTo)
    if apiSchemaCanOnlyApplyTo:
        primSpecCustomData['apiSchemaCanOnlyApplyTo'] = \
            Vt.TokenArray(apiSchemaCanOnlyApplyTo)

    if providesUsdShadeConnectableAPIBehavior:
        extraPlugInfo = {
            SchemaDefiningKeys.PROVIDES_USD_SHADE_CONNECTABLE_API_BEHAVIOR \
                    : True
        }
        for propKey in [SchemaDefiningKeys.IS_USD_SHADE_CONTAINER, \
                SchemaDefiningKeys.REQUIRES_USD_SHADE_ENCAPSULATION]:
            if propKey in sdrNodeMetadata:
                # Since we want to assign the types for these to bool and
                # because in python boolean type is a subset of int, we need to
                # do following instead of assign the propValue directly.
                propValue = StringToBool(sdrNodeMetadata[propKey])
                extraPlugInfo[propKey] = bool(propValue)

        primSpecCustomData['extraPlugInfo'] = extraPlugInfo

    primSpec.customData = primSpecCustomData

    doc = sdrNode.GetHelp()
    if doc != "":
        _SetSchemaUserDocFields(primSpec, doc)

    # Open schemaLayer on a stage so we can use the UsdUI hints API. We'll get
    # missing sublayer warnings for the base schemas since we're not configured
    # to resolve them, but these shouldn't matter since we only care about
    # authoring to schemaLayer. Suppress the warnings with a diagnostic
    # delegate.
    delegate = UsdUtils.CoalescingDiagnosticDelegate()
    stage = Usd.Stage.Open(schemaLayer)
    del delegate

    if not stage:
        Tf.RaiseCodingError(
            "Failed to open schema layer '%s' on a stage" \
            % schemaLayer.identifier)

    # Populate open-by-default display groups
    nodePrim = stage.GetPrimAtPath(primSpec.path)
    if not nodePrim:
        Tf.RaiseCodingError(
            "Failed to find composed node prim '%s'" % primSpec.name)

    primHints = UsdUI.PrimHints(nodePrim)
    for page in sdrNode.GetOpenPages():
        primHints.SetDisplayGroupExpanded(page, True)

    # gather properties from a prim definition generated by composing apiSchemas
    # provided by apiSchemasForAttrPruning metadata.
    primDefForAttrPruning = None
    if apiSchemasForAttrPruning:
        primDefForAttrPruning = usdSchemaReg.BuildComposedPrimDefinition(
                typedSchemaForAttrPruning, apiSchemasForAttrPruning)
    else:
        primDefForAttrPruning = \
            usdSchemaReg.FindConcretePrimDefinition(typedSchemaForAttrPruning)

    # We want the property order on the output schema to match the param order
    # from the source shader. SdrShaderNode should give the params to us in this
    # order, so remember it as we go.
    propOrder = []

    # `shownIf` expressions refer to shader params by their Sdr names. But as
    # we create attribute specs for each param, we modify their namespaces (see
    # _CreateAttrSpecFromNodeAttribute()).
    #
    # Track such changes so we can fix up shownIf expressions once the final
    # names are all known.
    attrsWithShownIf = []
    attrRenames = {}

    # Track "legacy" allowedTokens opinions (see note in _PopulateOptions()
    # above), so we can emit a warning below. Clients should move to consuming
    # valueLabels in these cases.
    attrsWithLegacyAllowedTokens = set()

    def _recordAttrInfo(attrName, sdrProp):
        # Property order
        propOrder.append(attrName)
        attrRenames[sdrProp.GetName()] = attrName

        # shownIf
        attr = stage.GetAttributeAtPath(
            primSpec.path.AppendProperty(attrName))
        attrHints = UsdUI.AttributeHints(attr)

        if not attrHints:
            return

        if attrHints.GetShownIf():
            attrsWithShownIf.append(attr)

        # Legacy allowedTokens
        attrSpec = primSpec.attributes.get(attrName)

        if not attrSpec:
            return

        if attrSpec.allowedTokens and attrHints.GetValueLabels():
            attrsWithLegacyAllowedTokens.add(attrName)

    # Create attrSpecs from input parameters
    for propName in sdrNode.GetShaderInputNames():
        sdrProp = sdrNode.GetShaderInput(propName)
        attrName = _CreateAttrSpecFromNodeAttribute(
                primSpec, sdrProp, stage,
                primDefForAttrPruning, schemaPropertyNSPrefixOverride)
        if attrName:
            _recordAttrInfo(attrName, sdrProp)

    # Create attrSpecs from output parameters
    # Note that we always want outputs: namespace prefix for output attributes.
    for propName in sdrNode.GetShaderOutputNames():
        sdrProp = sdrNode.GetShaderOutput(propName)
        attrName = _CreateAttrSpecFromNodeAttribute(
                primSpec, sdrProp, stage,
                primDefForAttrPruning, UsdShade.Tokens.outputs[:-1], False)
        if attrName:
            _recordAttrInfo(attrName, sdrProp)

    # Fix up both param-level and node-level shownIf expressions with the
    # final attribtue names, and write them
    for attr in attrsWithShownIf:
        attrHints = UsdUI.AttributeHints(attr)
        if attrHints:
            # XXX: Should be using Sdf.BooleanExpression.RenameVariables()
            # directly here, but it hasn't been wrapped yet
            attrHints.SetShownIf(
                UsdUtils._BooleanExpressionRenameVariables(
                    attrHints.GetShownIf(), attrRenames))

    pagesShownIf = sdrNode.GetPagesShownIf()
    for page, expr in pagesShownIf.items():
        # XXX: Ditto above
        pagesShownIf[page] = \
            UsdUtils._BooleanExpressionRenameVariables(expr, attrRenames)

    if pagesShownIf:
        primHints.SetDisplayGroupsShownIf(pagesShownIf)

    if attrsWithLegacyAllowedTokens:
        Tf.Warn("%s: Wrote both valueLabels and allowedTokens for %d "
                "attribute(s)" % \
                (sdrNode.GetName(), len(attrsWithLegacyAllowedTokens)))

    # Forward the source param order for typed and single-apply. Multiple-apply
    # schemas cannot specify propertyOrder.
    if (isTyped or isAPI) and propOrder:
        primSpec.propertyOrder = propOrder

    # Create token shaderId attrSpec -- only for shader nodes
    if (schemaBaseProvidesConnectability or \
            schemaPropertyNSPrefixOverride is None or \
            _IsNSPrefixConnectableAPICompliant(schemaPropertyNSPrefixOverride)):
        # We must add shaderId for all shaderNodes with the same identifier
        # across all sourceTypes, so that we get appropriate
        # renderContext:sourceType:shaderId attribute.
        sdrRegistry = Sdr.Registry()
        shaderNodesForShaderIdAttrs = [
            node for node in sdrRegistry.GetShaderNodesByIdentifier(
                sdrNode.GetIdentifier())]
        shaderIdAttrNames = set()
        for node in shaderNodesForShaderIdAttrs:

            shaderIdAttrNames.add(Sdf.Path.JoinIdentifier( \
                    [renderContext, node.GetContext(), 
                        PropertyDefiningKeys.SHADER_ID]))
        
        for shaderIdAttrName in shaderIdAttrNames:
            shaderIdAttrSpec = Sdf.AttributeSpec(primSpec, shaderIdAttrName,
                    Sdf.ValueTypeNames.Token, Sdf.VariabilityUniform)

            # Since users shouldn't need to be aware of shaderId attribute, we 
            # put this in "Internal" displayGroup.
            shaderIdAttrSpec.displayGroup = \
                    PropertyDefiningKeys.INTERNAL_DISPLAY_GROUP

            # We are iterating on sdrNodes which are guaranteed to be registered
            # with sdrRegistry and it only makes sense to add shaderId for these
            # shader nodes, so directly get the identifier from the node itself.
            shaderIdAttrSpec.default = sdrNode.GetIdentifier()

    # Extra attrSpec
    schemaBasePrimDefinition = \
        Usd.SchemaRegistry().FindConcretePrimDefinition(schemaBase)
    if schemaBasePrimDefinition and \
        SchemaDefiningMiscConstants.NodeDefAPI in \
        schemaBasePrimDefinition.GetAppliedAPISchemas():
            infoIdAttrSpec = Sdf.AttributeSpec(primSpec, \
                    UsdShade.Tokens.infoId, Sdf.ValueTypeNames.Token, \
                    Sdf.VariabilityUniform)
            infoIdAttrSpec.default = nodeIdentifier

    schemaLayer.Save()
